import numpy as np
import os
import multiprocessing as mp
from collections.abc import Iterable
from kliff.descriptors.descriptor import load_fingerprints
from kliff.error import InputError
from kliff.dataset import Configuration
import torch
import torch.nn as nn
from torch.utils.data import Dataset
from torch.utils.data import DataLoader

# pytorch buildin (use them directly)
from torch.nn.modules.linear import Linear, Bilinear
from torch.nn.modules.conv import (
    Conv1d,
    Conv2d,
    Conv3d,
    ConvTranspose1d,
    ConvTranspose2d,
    ConvTranspose3d,
)
from torch.nn.modules.activation import (
    Threshold,
    ReLU,
    Hardtanh,
    ReLU6,
    Sigmoid,
    Tanh,
    Softmax,
    Softmax2d,
    LogSoftmax,
    ELU,
    SELU,
    CELU,
    Hardshrink,
    LeakyReLU,
    LogSigmoid,
    Softplus,
    Softshrink,
    PReLU,
    Softsign,
    Softmin,
    Tanhshrink,
    RReLU,
    GLU,
)
from torch.nn.modules.pooling import (
    AvgPool1d,
    AvgPool2d,
    AvgPool3d,
    MaxPool1d,
    MaxPool2d,
    MaxPool3d,
    MaxUnpool1d,
    MaxUnpool2d,
    MaxUnpool3d,
    FractionalMaxPool2d,
    LPPool1d,
    LPPool2d,
    AdaptiveMaxPool1d,
    AdaptiveMaxPool2d,
    AdaptiveMaxPool3d,
    AdaptiveAvgPool1d,
    AdaptiveAvgPool2d,
    AdaptiveAvgPool3d,
)
from torch.nn.modules.batchnorm import BatchNorm1d, BatchNorm2d, BatchNorm3d
from torch.nn.modules.normalization import (
    LocalResponseNorm,
    CrossMapLRN2d,
    LayerNorm,
    GroupNorm,
)
from torch.nn.modules.rnn import (
    RNNBase,
    RNN,
    LSTM,
    GRU,
    RNNCellBase,
    RNNCell,
    LSTMCell,
    GRUCell,
)


@torch._jit_internal.weak_module
class Dropout(torch.nn.modules.dropout._DropoutNd):
    """A Dropout layer that zeros the same element of descriptor values for all
    atoms.

    Note `torch.nn.Dropout` dropout each component independently.

    Parameters
    ----------
    p: float
        probability of an element to be zeroed. Default: 0.5

    inplace: bool
        If set to `True`, will do this operation in-place. Default: `False`

    Shapes
    ------
        Input: [N, D] or [1, N, D]
        Outut: [N, D] or [1, N, D] (same as Input)
        The first dimension 1 is beause the dataloader provides only sample each
        iteration.
    """

    @torch._jit_internal.weak_script_method
    def forward(self, input):
        dim = input.dim()
        shape = input.shape
        if dim == 2:
            shape_4D = (1, *shape, 1)
        elif dim == 3:
            if shape[0] != 1:
                raise Exception('Shape[0] needs to be 1 for a 3D tensor.')
            shape_4D = (*shape, 1)

        else:
            raise Exception(
                'Input need to be 2D or 3D tensor, but got a ' '{}D tensor.'.format(dim)
            )
        x = torch.reshape(input, shape_4D)
        x = torch.transpose(x, 1, 2)
        y = torch.nn.functional.dropout2d(x, self.p, self.training, self.inplace)
        y = torch.transpose(y, 1, 2)
        y = torch.reshape(y, shape)
        return y


class FingerprintsDataset(Dataset):
    """Atomic environment fingerprints dataset.

    Parameters
    ----------
    fname: string
        Name of the fingerprints file.

    transform: callable (optional):
        Optional transform to be applied on a sample.
    """

    def __init__(self, fname, transform=None):
        self.fp = load_fingerprints(fname)
        self.transform = transform

    def __len__(self):
        return len(self.fp)

    def __getitem__(self, index):
        sample = self.fp[index]
        if self.transform:
            sample = self.transform(sample)
        return sample


# TODO implement here GPU options
class FingerprintsDataLoader(DataLoader):
    """A dataset loader that incorporate the support the number of epochs.

    The dataset loader will load an element from the next batch if a batch is
    fully iterarated. This, in effect, looks like concatenating the dataset the
    number of epochs times.

    Parameters
    ----------
    num_epochs: int
        Number of epochs to iterate through the dataset.
    """

    def __init__(self, num_epochs=1, *args, **kwargs):
        super(FingerprintsDataLoader, self).__init__(*args, **kwargs)
        self.num_epochs = num_epochs
        self.epoch = 0
        self.iterable = None

    def next_element(self):
        """ Get the next data element.
        """
        if self.iterable is None:
            self.iterable = self._make_iterable()
        try:
            element = self.iterable.next()
        except StopIteration:
            self.epoch += 1
            if self.epoch == self.num_epochs:
                raise StopIteration
            else:
                self.iterable = self._make_iterable()
                element = self.next_element()
        return element

    def _make_iterable(self):
        iterable = iter(self)
        return iterable


class NeuralNetwork(nn.Module):
    """Neural Network model.

    A feed-forward neural network model built using PyTorch.

    Parameters
    ----------
    descriptor: object
        A descriptor that transforms atomic environment information to the
        fingerprints, which are used as the input for the neural network.

    seed: int (optional)
        Global seed for random numbers.
    """

    def __init__(self, descriptor, seed=35):
        super(NeuralNetwork, self).__init__()
        self.descriptor = descriptor
        self.seed = seed

        dtype = self.descriptor.get_dtype()
        if dtype == np.float32:
            self.dtype = torch.float32
        elif dtype == np.float64:
            self.dtype = torch.float64
        else:
            raise NeuralNetworkError('Not support dtype "{}".'.format(dtype))

        torch.manual_seed(seed)
        self.layers = None
        self.save_prefix = None
        self.save_start = None
        self.save_frequency = None

    def add_layers(self, *layers):
        """Add layers to the sequential model.

        Parameters
        ----------
        layers: torch.nn layers
            ``torch.nn`` layers that are used to build a sequential model.
            Available ones including: torch.nn.Linear, torch.nn.Dropout, and
            torch.nn.Sigmoid among others. See
            https://pytorch.org/docs/stable/nn.html
            for a full list of torch.nn layers.
        """
        if self.layers is not None:
            raise NeuralNetworkError(
                '"add_layers" called multiple times. ' 'It should be called only once.'
            )
        else:
            self.layers = []

        for la in layers:
            self.layers.append(la)
            # set it as attr so that parameters are automatically registered
            setattr(self, 'layer_{}'.format(len(self.layers)), la)

        # check shape of first layer and last layer
        first = self.layers[0]
        if first.in_features != len(self.descriptor):
            raise InputError(
                '"in_features" of first layer should be equal to ' 'descriptor size.'
            )
        last = self.layers[-1]
        if last.out_features != 1:
            raise InputError('"out_features" of last layer should be 1.')

        # cast types
        self.type(self.dtype)

    def forward(self, x):
        for j, layer in enumerate(self.layers):
            x = layer(x)
        return x

    def set_save_metadata(self, prefix, start, frequency):
        """Set metadata that controls how the model are saved during training.

        If this function is called before minimization starts, the model will be
        saved to the directory specified by ``prefix`` every ``frequency``
        epochs, beginning at the ``start`` epoch.

        Parameters
        ----------
        prefix: str
            Directory where the models are saved.
            Model will be named as `{prefix}/model_epoch{ep}.pt`, where `ep` is
            the epoch number.

        start: int
            Eopch number at which begins to save the model.

        frequency: int
            Save the model every ``frequency`` epochs.
        """
        self.save_prefix = str(prefix)
        self.save_start = int(start)
        self.save_frequency = int(frequency)

    def save(self, path):
        """Save a model to disk.

        Parameters
        ----------
        path: str
            Path where to store the model.
        """
        dirname = os.path.dirname(path)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)
        torch.save(self.state_dict(), path)

    def load(self, path, mode):
        """Load a model on disk into memory.

        Parameters
        ----------
        path: str
            Path where the model is stored.

        mode: str
            Purpose of the loaded model. Should be either ``train`` or ``eval``.
        """

        self.load_state_dict(torch.load(path))
        if mode == 'train':
            self.train()
        elif mode == 'eval':
            self.eval()
        else:
            raise NeuralNetworkError(
                'Uncongnized mode "{}" in model.load().'.format(mode)
            )

    def write_kim_model(self, path=None):
        # supported
        param_layer = ['Linear']
        activ_layer = ['Sigmoid', 'Tanh', 'ReLU', 'ELU']
        dropout_layer = ['Dropout']
        layer_groups = self._group_layers(param_layer, activ_layer, dropout_layer)

        weights, biases = self._get_weights_and_biases(layer_groups, param_layer)
        activations = self._get_activations(layer_groups, activ_layer)
        drop_ratios = self._get_drop_ratios(layer_groups, dropout_layer)

        descriptor = self.descriptor
        dtype = self.dtype

        if path is None:
            path = os.path.join(os.getcwd(), 'NeuralNetwork__MO_000000111111_000')
        if path and not os.path.exists(path):
            os.makedirs(path)

        # TODO write CMakeList and probably pull the Eigen library
        with open(os.path.join(path, 'CMakeLists.txt'), 'w') as fout:
            fout.write('to be filled')

        # write parameter file
        fname = 'kliff_trained.params'
        with open(os.path.join(path, fname), 'w') as fout:
            fout.write('#' + '=' * 80 + '\n')
            fout.write(
                '# KIM ANN potential parameters, generated by `kliff` '
                'fitting program.\n'
            )
            fout.write('#' + '=' * 80 + '\n\n')

            # cutoff
            cutname, rcut = descriptor.get_cutoff()
            maxrcut = max(rcut.values())

            fout.write('# cutoff    rcut\n')
            if dtype == torch.float64:
                fout.write('{}  {:.15g}\n\n'.format(cutname, maxrcut))
            else:
                fout.write('{}  {:.7g}\n\n'.format(cutname, maxrcut))

            # symmetry functions
            # header
            fout.write('#' + '=' * 80 + '\n')
            fout.write('# symmetry functions\n')
            fout.write('#' + '=' * 80 + '\n\n')

            desc = descriptor.get_hyperparams()
            num_desc = len(desc)
            fout.write('{}    #number of symmetry funtion types\n\n'.format(num_desc))

            # descriptor values
            fout.write('# sym_function    rows    cols\n')
            for name, values in desc.items():
                if name == 'g1':
                    fout.write('g1\n\n')
                else:
                    rows = len(values)
                    cols = len(values[0])
                    fout.write('{}    {}    {}\n'.format(name, rows, cols))
                    if name == 'g2':
                        for val in values:
                            if dtype == torch.float64:
                                fout.write('{:.15g} {:.15g}'.format(val[0], val[1]))
                            else:
                                fout.write('{:.7g} {:.7g}'.format(val[0], val[1]))
                            fout.write('    # eta  Rs\n')
                        fout.write('\n')
                    elif name == 'g3':
                        for val in values:
                            if dtype == torch.float64:
                                fout.write('{:.15g}'.format(val[0]))
                            else:
                                fout.write('{:.7g}'.format(val[0]))
                            fout.write('    # kappa\n')
                        fout.write('\n')
                    elif name == 'g4':
                        for val in values:
                            zeta = val[0]
                            lam = val[1]
                            eta = val[2]
                            if dtype == torch.float64:
                                fout.write(
                                    '{:.15g} {:.15g} {:.15g}'.format(zeta, lam, eta)
                                )
                            else:
                                fout.write('{:.7g} {:.7g} {:.7g}'.format(zeta, lam, eta))
                            fout.write('    # zeta  lambda  eta\n')
                        fout.write('\n')
                    elif name == 'g5':
                        for val in values:
                            zeta = val[0]
                            lam = val[1]
                            eta = val[2]
                            if dtype == torch.float64:
                                fout.write(
                                    '{:.15g} {:.15g} {:.15g}'.format(zeta, lam, eta)
                                )
                            else:
                                fout.write('{:.7g} {:.7g} {:.7g}'.format(zeta, lam, eta))
                            fout.write('    # zeta  lambda  eta\n')
                        fout.write('\n')

            # data centering and normalization
            # header
            fout.write('#' + '=' * 80 + '\n')
            fout.write('# Preprocessing data to center and normalize\n')
            fout.write('#' + '=' * 80 + '\n')

            # mean and stdev
            mean = descriptor.get_mean()
            stdev = descriptor.get_stdev()
            if mean is None and stdev is None:
                fout.write('center_and_normalize  False\n')
            else:
                fout.write('center_and_normalize  True\n\n')

                fout.write('# mean\n')
                for i in mean:
                    if dtype == torch.float64:
                        fout.write('{:23.15e}\n'.format(i))
                    else:
                        fout.write('{:15.7e}\n'.format(i))
                fout.write('\n# standard deviation\n')
                for i in stdev:
                    if dtype == torch.float64:
                        fout.write('{:23.15e}\n'.format(i))
                    else:
                        fout.write('{:15.7e}\n'.format(i))
                fout.write('\n')

            # ann structure and parameters
            # header
            fout.write('#' + '=' * 80 + '\n')
            fout.write('# ANN structure and parameters\n')
            fout.write('#\n')
            fout.write(
                '# Note that the ANN assumes each row of the input "X" '
                'is an observation, i.e.\n'
            )
            fout.write('# the layer is implemented as\n')
            fout.write('# Y = activation(XW + b).\n')
            fout.write(
                '# You need to transpose your weight matrix if each '
                'column of "X" is an observation.\n'
            )
            fout.write('#' + '=' * 80 + '\n\n')

            # number of layers
            num_layers = len(weights)
            fout.write(
                '{}    # number of layers (excluding input layer, '
                'including output layer)\n'.format(num_layers)
            )

            # size of layers
            for b in biases:
                fout.write('{}  '.format(len(b)))
            fout.write('  # size of each layer (last must be 1)\n')

            # activation function
            # TODO enable writing different activations for each layer
            activation = activations[0]
            fout.write('{}    # activation function\n'.format(activation))

            # keep probability
            for i in drop_ratios:
                fout.write('{:.15g}  '.format(1.0 - i))
            fout.write('  # keep probability of input for each layer\n\n')

            # weights and biases
            for i, (w, b) in enumerate(zip(weights, biases)):

                # weight
                rows, cols = w.shape
                if i != num_layers - 1:
                    fout.write(
                        '# weight of hidden layer {} (shape({}, {}))\n'.format(
                            i + 1, rows, cols
                        )
                    )
                else:
                    fout.write(
                        '# weight of output layer (shape({}, {}))\n'.format(rows, cols)
                    )
                for line in w:
                    for item in line:
                        if dtype == torch.float64:
                            fout.write('{:23.15e}'.format(item))
                        else:
                            fout.write('{:15.7e}'.format(item))
                    fout.write('\n')

                # bias
                if i != num_layers - 1:
                    fout.write(
                        '# bias of hidden layer {} (shape({}, {}))\n'.format(
                            i + 1, rows, cols
                        )
                    )
                else:
                    fout.write(
                        '# bias of output layer (shape({}, {}))\n'.format(rows, cols)
                    )
                for item in b:
                    if dtype == torch.float64:
                        fout.write('{:23.15e}'.format(item))
                    else:
                        fout.write('{:15.7e}'.format(item))
                fout.write('\n\n')

    def _group_layers(self, param_layer, activ_layer, dropout_layer):
        """Divide all the layers into groups.

        The first group is either an empty list or a `Dropout` layer for the
        input layer. The last group typically contains only a `Linear` layer.
        For other groups, each group contains two, or three layers. `Linear`
        layer and an activation layer are mandatory, and a third `Dropout` layer
        is optional.

        Return
        ------
        groups: list of list of layers
        """

        groups = []
        new_group = []

        supported = param_layer + activ_layer + dropout_layer
        for i, layer in enumerate(self.layers):
            name = layer.__class__.__name__
            if name not in supported:
                raise NeuralNetworkError(
                    'Layer "{}" not supported by KIM model. Cannot proceed '
                    'to write.'.format(name)
                )
            if name in activ_layer:
                if i == 0:
                    raise NeuralNetworkError(
                        'First layer cannot be a "{}" layer'.format(name)
                    )
                if self.layers[i - 1].__class__.__name__ not in param_layer:
                    raise NeuralNetworkError(
                        'Cannot convert to KIM model. a "{}" layer must follow '
                        'a "Linear" layer.'.format(name)
                    )
            if name[:7] in dropout_layer:
                if self.layers[i - 1].__class__.__name__ not in activ_layer:
                    raise NeuralNetworkError(
                        'Cannot convert to KIM model. a "{}" layer must follow '
                        'an activation layer.'.format(name)
                    )
            if name in param_layer:
                groups.append(new_group)
                new_group = []
            new_group.append(layer)
        groups.append(new_group)

        return groups

    @staticmethod
    def _get_weights_and_biases(groups, supported):
        """Get weights and biases of all layers that have weights and biases.
        """
        weights = []
        biases = []
        for i, g in enumerate(groups):
            if i != 0:
                layer = g[0]
                name = layer.__class__.__name__
                if name in supported:
                    weight = layer.weight
                    bias = layer.bias
                    weights.append(weight)
                    biases.append(bias)
        return weights, biases

    @staticmethod
    def _get_activations(groups, supported):
        """Get the activation of all layers."""
        activations = []
        for i, g in enumerate(groups):
            if i != 0 and i != (len(groups) - 1):
                layer = g[1]
                name = layer.__class__.__name__
                if name in supported:
                    activations.append(name.lower())
        return activations

    @staticmethod
    def _get_drop_ratios(groups, supported):
        """Get the dropout ratio of all layers."""
        drop_ratios = []
        for i, g in enumerate(groups):
            if i == 0:
                if len(g) != 0:
                    layer = g[0]
                    name = layer.__class__.__name__
                    if name in supported:
                        drop_ratios.append(layer.p)
                else:
                    drop_ratios.append(0.0)
            elif i == len(groups) - 1:
                pass
            else:
                if len(g) == 3:
                    layer = g[2]
                    name = layer.__class__.__name__
                    if name in supported:
                        drop_ratios.append(layer.p)
                else:
                    drop_ratios.append(0.0)

        return drop_ratios


class PytorchANNCalculator(object):
    """ A neural network calculator.

    Parameters
    ----------
    model: obj
        Instance of :class:`~kliff.neuralnetwork.NeuralNetwork`.

    Attributes
    ----------
    attr1: list
        This is an example attribute.
    """

    # TODO should be moved to Model
    implemented_property = ['energy', 'forces']

    def __init__(self, model):

        self.model = model
        self.dtype = self.model.descriptor.dtype
        self.train_fingerprints_path = None

        self.use_energy = None
        self.use_forces = None

        self.results = dict([(i, None) for i in self.implemented_property])

    def create(
        self,
        configs,
        use_energy=True,
        use_forces=True,
        use_stress=False,
        reuse=False,
        nprocs=mp.cpu_count(),
    ):
        """Preprocess configs into fingerprints.

        Parameters
        ----------

        configs: list of Configuration object

        use_energy: bool (optional)
            Whether to require the calculator to compute energy.

        use_forces: bool (optional)
            Whether to require the calculator to compute forces.

        use_stress: bool (optional)
            Whether to require the calculator to compute stress.

        nprocs: int (optional)
            Number if processors.

        """
        if use_stress:
            raise NotImplementedError('"stress" is not supported by NN calculator.')

        self.configs = configs
        self.use_energy = use_energy
        self.use_forces = use_forces

        if isinstance(configs, Configuration):
            configs = [configs]

        # generate pickled fingerprints
        print('Start generating fingerprints')
        fname = self.model.descriptor.generate_train_fingerprints(
            configs, grad=use_forces, reuse=reuse, nprocs=nprocs
        )
        print('Finish generating fingerprints')
        self.train_fingerprints_path = fname

    def get_train_fingerprints_path(self):
        """Return the path to the training set fingerprints: `train.pkl`."""
        return self.train_fingerprints_path

    def compute(self, x):

        grad = self.use_forces
        zeta = x['zeta'][0]

        if grad:
            zeta.requires_grad = True
        y = self.model(zeta)
        pred_energy = y.sum()
        if grad:
            dzeta_dr = x['dzeta_dr'][0]
            forces = self.compute_forces(pred_energy, zeta, dzeta_dr)
            zeta.requires_grad = False
        else:
            forces = None

        return {'energy': pred_energy, 'forces': forces}

    @staticmethod
    def compute_forces(energy, zeta, dzeta_dr):
        denergy_dzeta = torch.autograd.grad(energy, zeta, create_graph=True)[0]
        forces = -torch.tensordot(denergy_dzeta, dzeta_dr, dims=([0, 1], [0, 1]))
        return forces


def cost_single_config(pred_energy, energy=None, forces=None):
    cost = (pred_energy - energy) ** 2
    return cost


class NeuralNetworkError(Exception):
    def __init__(self, msg):
        super(NeuralNetworkError, self).__init__(msg)
        self.msg = msg

    def __expr__(self):
        return self.msg



"""
.. _tut_nn:

Train a neural network potential
================================

In this tutorial, we train a neural network (NN) potential for silicon

"""


##########################################################################################
# We are going to fit the NN potential to a training set of energies and forces from
# compressed and stretched diamond silicon structures (the same training set used in
# :ref:`tut_kim_sw`).
# Download the training set :download:`Si_training_set.tar.gz <https://raw.githubusercontent.com/mjwen/kliff/pytorch/examples/Si_training_set.tar.gz>`
# and extract the tarball: ``$ tar xzf Si_training_set.tar.gz``.
# The data is stored in **extended xyz** format, and see :ref:`doc.dataset` for more
# information of this format.
#
# .. warning::
#     The ``Si_training_set`` is just a toy data set for the purpose to demonstrate how to
#     use KLIFF to train potentials. It should not be used to train any potential for real
#     simulations.
#
# Let's first import the modules that will be used in this example.

from kliff.descriptors import SymmetryFunction
from kliff.dataset import DataSet
import kliff.neuralnetwork as nn
from kliff.loss import Loss


##########################################################################################
# Model
# -----
#
# For a NN model, we need to specify the descriptor that transforms atomic environment
# information to the fingerprints, which the NN modle uses as the input. Here, we use the
# symmetry functions proposed by by Behler and coworkers.

descriptor = SymmetryFunction(
    cut_name='cos', cut_dists={'Si-Si': 5.0}, hyperparams='set31', normalize=True
)


##########################################################################################
# The ``cut_name`` and ``cut_dists`` tells the descriptor what type of cutoff function to
# use and what the cutoff distances are. ``hyperparams`` specifies the the set of
# hyperparameters used in the symmetry function descriptor. If you prefer, you can provide
# a dictionary of your own hyperparameters. And finally, ``normalize`` informs that the
# genereated fingerprints should be normalized by first subtracting the mean and then
# dividing the standard deviation. This normalization typically makes it easier to
# optimzie  NN model.
#
# We can then build the NN model on top of the descriptor.

N1 = 10
N2 = 10
model = nn.NeuralNetwork(descriptor)
model.add_layers(
    # first hidden layer
    nn.Linear(descriptor.get_size(), N1),
    nn.Tanh(),
    # second hidden layer
    nn.Linear(N1, N2),
    nn.Tanh(),
    # output layer
    nn.Linear(N2, 1),
)
model.set_save_metadata(prefix='./my_kliff_model', start=5, frequency=2)


##########################################################################################
# In the above code, we build a NN model with an input layer, two hidden layer, and an
# output layer. The ``descriptor`` carries the information of the input layer, so it is
# not needed to be spcified explicitly. For each hidden layer, we first do a linear
# transformation using ``nn.Linear(size_in, size_out)`` (essentially carrying out :math:`y
# = xW+b`, where :math:`W` is the weight matrix of size ``size_in`` by ``size_out``, and
# :math:`b` is a vector of size ``size_out``. Then we apply the hyperbolic tangent
# activation function ``nn.Tanh()`` to the output of the Linear layer (i.e. :math:`y`) so
# as to add the nonlinearty. We use a Linear layer for the output layer as well, but
# unlike the hidden layer, no activation function is applied here. The input size
# ``size_in`` of the first hidden layer must be the size of the descriptor, which is
# obtained using ``descriptor.get_size()``. For all other layers (hidden or output), the
# input size must be equal to the output size of the previous layer. The ``out_size`` of
# the output layer much be 1 such that the output of the NN model is gives the energy of
# atom.
#
# The ``set_save_metadata`` function call informs where to save intermediate models during
# the optimization (discussed below), and what the starting epoch and how often to save
# the model.
#
#
# Training set and calculator
# ---------------------------
#
# The training set and the calculator are the same as explaned in :ref:`tut_kim_sw`. The
# only difference is that we need use the
# :mod:`~kliff.neuralnetwork.PytorchANNCalculator()`, which is targeted for the NN model.
# Also, its ``create()`` method takes an argument ``reuse`` to inform whether to reuse the
# fingerprints generated from the descriptor if it is present.

# training set
dataset_name = 'Si_training_set/varying_alat'
tset = DataSet()
tset.read(dataset_name)
configs = tset.get_configs()
print('Number of configurations:', len(configs))

# calculator
calc = nn.PytorchANNCalculator(model)
calc.create(configs, reuse=True)


##########################################################################################
# Loss function
# -------------
#
# KLIFF uses a loss function to quantify the difference between the training data and
# potential predictions and uses minimization algorithms to reduce the loss as much as
# possible. In the following code snippet, we create a loss function that uses the
# ``Adam`` optimzier to minimize it. The Adam optimizer supports minimization using
# `mini-batches` of data, and here we use ``100`` configurations in each minimization step
# (the training set has a total of 400 configuraions as can be seen above), and run
# through the training set for ``10`` epochs. The learning rate ``lr`` used here is
# ``0.01``, and typically, one may need to play with this to find an acceptable one that
# drives the loss down in a reasonable time.

loss = Loss(calc, residual_data={'forces_weight': 0.3})
result = loss.minimize(method='Adam', num_epochs=10, batch_size=100, lr=0.01)


##########################################################################################
# We can save the trained model to disk, and later can load it back if we want. We can
# also write the trained model to a KIM model such that it can be used in other simulation
# codes such as LAMMPS via the KIM API.

model.save('./saved_model.pkl')
model.write_kim_model()



# Training a model using multiple processes:
import torch
import torch.nn as nn
import torch.multiprocessing as mp
import numpy as np
from torch.utils.data import TensorDataset, DataLoader

torch.manual_seed(35)


def init_processes(rank, size, fn, backend='tcp'):
    """ Initialize the distributed environment. """
    os.environ['MASTER_ADDR'] = '127.0.0.1'
    os.environ['MASTER_PORT'] = '29500'
    dist.init_process_group(backend, rank=rank, world_size=size)
    fn(rank, size)


def loss_fn(pref, label):
    return torch.sum(torch.pow(pref - label, 2))


def train(model, tset):
    optimizer = torch.optim.Adam(model.parameters(), lr=0.01)
    for data, labels in tset:
        print('@@lable', labels)
        optimizer.zero_grad()
        loss_fn(model(data), labels).backward()
        optimizer.step()  # This will update the shared parameters


def create_dataset():
    X = torch.empty(20, 5).uniform_(0, 1)
    Y = np.asarray(np.atleast_2d(0.1 * np.arange(20)).T, dtype=np.float32)
    Y = torch.from_numpy(Y)
    dataset = TensorDataset(X, Y)
    dataset = DataLoader(dataset, batch_size=4, shuffle=True, num_workers=0)
    return dataset


n_in = 5
n_h1 = 10
n_out = 1
model = nn.Sequential(nn.Linear(n_in, n_h1), nn.ReLU(), nn.Linear(n_h1, n_out))
model.share_memory()  # Required for 'fork' method to work

print('@@ params before:', model.parameters())
for i in model.parameters():
    print(i)
    break

init_processes()
dataset = create_dataset()
torch.distributed.init_process_group(backend='gloo')
model = torch.nn.DistributedDataParallelCPU(model)
train(model, dataset)


print('@@ params after:', model.parameters())
for i in model.parameters():
    print(i)
    break



import multiprocessing as mp
import random
import numpy as np


def parmap1(f, X, *args, tuple_X=False, nprocs=mp.cpu_count()):
    """Parallelism over data.

    This function mimics ``multiprocessing.Pool.map`` to allow extra arguments
    to be used for the function ``f``.

    Parameters
    ----------
    f: function
        The function that operates on the data.

    X: list
        Data to be parallelized.

    args: args
        Extra positional arguments needed by the function ``f``.

    tuple_X: bool
        This depends on ``X``. It should be set to ``True`` if multiple
        arguments are parallelized and set to ``False`` if only one argument is
        parallelized. See ``Example`` below.

    nprocs: int
        Number of processors to use.

    Return
    ------
    list
        A list of results, corresponding to ``X``.

    Note
    ----
    The data is put into a job queue, a worker process gets a piece of the data
    to work on, the worker pushes the result back to the manager through another
    queue, and then get another piece of data until the job queue is empty.
    So, in principle, there will not be idle worker and it should be faster than
    :meth:`kliff.parallel.parmap2`.

    Warning
    -------
    This is implemented using ``multiprocessing.Queue``, which requires the
    funciton ``f`` to be picklable. If it is not the case (e.g. use KIM library
    functions), use :meth:`kliff.parallel.parmap2` that is based on
    ``multiprocessing.Pipe``.


    Example
    -------
    >>> def func(x, y, z=1):
    >>>     return x+y+z
    >>> X = range(3)
    >>> Y = range(3)
    >>> parmap1(func, X, 1, nprocs=2)  # [2,3,4]
    >>> parmap1(func, X, 1, 1, nprocs=2)  # [2,3,4]
    >>> parmap1(func, zip(X, Y), tuple_X=True, nprocs=2)  # [1,3,5]
    >>> parmap1(func, zip(X, Y), 1, tuple_X=True, nprocs=2)  # [1,3,5]
    """

    q_in = mp.Queue(nprocs)
    q_out = mp.Queue()

    processes = []
    for _ in range(nprocs):
        p = mp.Process(target=_func1, args=(f, q_in, q_out))
        p.daemon = True
        p.start()
        processes.append(p)

    N = 0
    for i, x in enumerate(X):
        N += 1
        if tuple_X:
            ix = (i, *x)
        else:
            ix = (i, x)
        q_in.put((ix, args))

    [q_in.put((None, None)) for _ in range(nprocs)]
    results = [q_out.get() for _ in range(N)]
    [p.join() for p in processes]

    return [r for i, r in sorted(results)]


def _func1(f, q_in, q_out):
    while True:
        ix, args = q_in.get()
        if ix is None:
            break
        i = ix[0]
        x = ix[1:]
        y = f(*x, *args)
        q_out.put((i, y))


def parmap2(f, X, *args, tuple_X=False, nprocs=mp.cpu_count()):
    """Parallelism over data.

    This is to mimic ``multiprocessing.Pool.map``, which requires the function
    ``f`` to be picklable. This function does not have this restriction and
    allows extra arguments to be used for the function ``f``.

    Parameters
    ----------
    f: function
        The function that operates on the data.

    X: list
        Data to be parallelized.

    args: args
        Extra positional arguments needed by the function ``f``.

    tuple_X: bool
        This depends on ``X``. It should be set to ``True`` if multiple
        arguments are parallelized and set to ``False`` if only one argument is
        parallelized. See ``Example`` below.

    nprocs: int
        Number of processors to use.

    Return
    ------
    list
        A list of results, corresponding to ``X``.

    Note
    ----
    This function is implemented using ``multiprocessing.Pipe``. The data is
    subdivided into ``nprocs`` groups and then each group of data is distributed
    to a process. The results from each group are then assembled together.
    The data is shuffled to balance the load in each process.
    See :meth:`kliff.parallel.parmap1` for another implementation that uses
    ``multiprocessing.Queue``.

    Example
    -------
    >>> def func(x, y, z=1):
    >>>     return x+y+z
    >>> X = range(3)
    >>> Y = range(3)
    >>> parmap2(func, X, 1, nprocs=2)  # [2,3,4]
    >>> parmap2(func, X, 1, 1, nprocs=2)  # [2,3,4]
    >>> parmap2(func, zip(X, Y), tuple_X=True, nprocs=2)  # [1,3,5]
    >>> parmap2(func, zip(X, Y), 1, tuple_X=True, nprocs=2)  # [1,3,5]
    """

    # shuffle and divide into nprocs equally-numbered parts
    # TODO this could be improved by schedule job according to number of atoms
    if tuple_X:
        pairs = [(i, *x) for i, x in enumerate(X)]  # to make array_split work
    else:
        pairs = [(i, x) for i, x in enumerate(X)]
    random.shuffle(pairs)
    groups = np.array_split(pairs, nprocs)

    processes = []
    managers = []
    for i in range(nprocs):
        manager_end, worker_end = mp.Pipe(duplex=False)
        p = mp.Process(target=_func3, args=(f, groups[i], args, worker_end))
        p.daemon = True
        p.start()
        processes.append(p)
        managers.append(manager_end)
    results = []
    for m in managers:
        results.extend(m.recv())
    for p in processes:
        p.join()

    return [r for i, r in sorted(results)]


def _func3(f, iX, args, worker_end):
    results = []
    for ix in iX:
        i = ix[0]
        x = ix[1:]
        results.append((i, f(*x, *args)))
    worker_end.send(results)



import logging


class Logger(object):
    def __init__(self, level='warning', filename='kliff.log'):
        self.level = level.upper()
        self.filename = filename
        logging.basicConfig(
            filename=self.filename,
            format='%(asctime)s:%(name)s:%(levelname)s: %(message)s',
        )
        self.loggers = []

    def set_level(self, level):
        self.level = level.upper()
        # reset level for previous instantiated loggers
        for l in self.loggers:
            l.setLevel(self.level)

    def get_logger(self, name=None):
        if name is None:
            logger = logging.getLogger()
        else:
            logger = logging.getLogger(name)
        logger.setLevel(self.level)
        self.loggers.append(logger)
        return logger



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
import copy
import sys


class Fisher:
    """Fisher information matrix.

    Compute the Fisher information according to $I_{ij} = \sum_m df_m/dp_i * df_m/dp_j$,
    where f_m are the forces on atoms in configuration m, p_i is the ith model parameter.
    Derivatives are computed numerically using Ridders' algorithm.

    Parameters
    ----------

    KIMobjs: list of KIMcalculator objects

    params: ModelParameters object
    """

    def __init__(self, params, calculator):
        self.params = params
        self.calculator = calculator
        self.F = None
        self.F_std = None
        self.delta_params = None

    def compute(self):
        """Comptue the Fisher information matrix and the standard deviation.

        Returns
        -------

        F: 2D array, shape(N, N), where N is the number of parameters
          Fisher informaiton matrix (FIM)

        F_std: 2D array, shape(N, N), where N is the number of parameters
          standard deviation of FIM
        """
        F_all = []
        kim_in_out_data = self.calculator.get_kim_input_and_output()
        for in_out in kim_in_out_data:
            dfdp = self._get_derivative_one_conf(in_out)
            F_all.append(np.dot(dfdp, dfdp.T))
        self.F = np.mean(F_all, axis=0)
        self.F_std = np.std(F_all, axis=0)
        return self.F, self.F_std

    def _get_derivative_one_conf(self, in_out):
        """Compute the derivative dfm/dpi for one atom configuration.

        Parameters
        ----------

        in_out: Configuration object
        """
        try:
            import numdifftools as nd
        except ImportError as e:
            raise ImportError(
                str(e) + '.\nFisher information computation needs '
                '"numdifftools". Please install first.'
            )

        derivs = []
        ori_param_vals = self.params.get_x0()
        for i, p in enumerate(ori_param_vals):
            values = copy.deepcopy(ori_param_vals)
            Jfunc = nd.Jacobian(self._get_prediction)
            df = Jfunc(p, i, values, in_out)
            derivs.append(df.reshape((-1,)))
            # restore param values back
            self.params.update_params(ori_param_vals)

        return np.array(derivs)

    def _get_prediction(self, x, idx, values, in_out):
        """ Compute predictions using specific parameter.

        Parameters
        ----------

        values: list of float
          the parameter values

        idx: int
          the index of 'x' in the value list

        x: float
          the specific parameter value at slot 'idx'

        Return
        ------
        forces: list of floats
          the forces on atoms in this configuration
        """
        values[idx] = x
        self.params.update_params(values)
        self.calculator.update_params(self.params)
        self.calculator.compute(in_out)
        forces = self.calculator.get_forces(in_out)
        forces = np.reshape(forces, (-1,))
        return forces



def error_report(self, normalize_by_num_atoms=True, fname='ERROR_REPORT'):
    """Write error of each configuration to fname.

        Parameters
        ----------
        fname: str
          path to the file to write the error report.
        """

    loss = self.get_loss(self.calculator.get_opt_params())

    with open(fname, 'w') as fout:
        fout.write('\n' + '=' * 80 + '\n')
        fout.write('Final loss: {:18.10e}\n\n'.format(loss))
        fout.write('=' * 80 + '\n')
        if normalize_by_num_atoms:
            fout.write(
                '(Loss, energy RMSE, and forces RMSE are normalized by '
                'number of atoms: Natoms.)\n\n'
            )
        fout.write(
            '      Loss       energy RMSE     forces RMSE  Natoms  '
            'config. identifier\n\n'
        )

        cas = self.calculator.get_compute_arguments()
        for ca in cas:

            # prediction data
            self.calculator.compute(ca)
            pred = self.calculator.get_prediction(ca)

            # reference data
            ref = self.calculator.get_reference(ca)

            conf = ca.conf
            identifier = conf.get_identifier()
            natoms = conf.get_number_of_atoms()

            compute_energy = ca.get_compute_flag('energy')
            compute_forces = ca.get_compute_flag('forces')
            if compute_energy:
                pred_energy = self.calculator.get_energy(ca)
                ref_energy = ca.conf.get_energy()
                energy_rmse = pred_energy - ref_energy
            else:
                energy_rmse = None
            if compute_forces:
                pred_forces = self.calculator.get_forces(ca)
                ref_forces = ca.conf.get_forces()
                forces_rmse = np.linalg.norm(pred_forces - ref_forces)
            else:
                forces_rmse = None

            residual = self.residual_fn(identifier, natoms, pred, ref, self.residual_data)
            loss = 0.5 * np.linalg.norm(residual) ** 2

            if normalize_by_num_atoms:
                nz = natoms
            else:
                nz = 1

            if energy_rmse is None:
                if forces_rmse is None:
                    fout.write(
                        '{:14.6e} {} {}   {}   {}\n'.format(
                            loss / nz,
                            energy_rmse / nz,
                            forces_rmse / nz,
                            natoms,
                            identifier,
                        )
                    )
                else:
                    fout.write(
                        '{:14.6e} {} {:14.6e}   {}   {}\n'.format(
                            loss / nz,
                            energy_rmse / nz,
                            forces_rmse / nz,
                            natoms,
                            identifier,
                        )
                    )
            else:
                if forces_rmse is None:
                    fout.write(
                        '{:14.6e} {:14.6e} {}   {}   {}\n'.format(
                            loss / nz,
                            energy_rmse / nz,
                            forces_rmse / nz,
                            natoms,
                            identifier,
                        )
                    )
                else:
                    fout.write(
                        '{:14.6e} {:14.6e} {:14.6e}   {}   {}\n'.format(
                            loss / nz,
                            energy_rmse / nz,
                            forces_rmse / nz,
                            natoms,
                            identifier,
                        )
                    )



from fisher import Fisher
from kimcalculator import KIMcalculator
from modelparams import ModelParams
from dataset import DataSet
from __future__ import print_function
import numpy as np
import sys

sys.path.append('../openkim_fit')


def test_fisher(relative_variance=False):

    # temperature
    T = 150

    # KIM model parameters
    modelname = 'Three_Body_Stillinger_Weber_MoS__MO_000000111111_000'
    params = ModelParams(modelname)
    params.echo_avail_params()
    fname = 'input/fitted_params_T150.txt'
    # fname = '/media/sf_share/mos2_fitted-T{}_interval4'.format(T)
    params.read(fname)
    params.echo_params()

    # read config and reference data
    tset = DataSet()
    fname = 'training_set/training_set_multi_small'
    # fname = 'training_set/training_set_multi_large'
    # fname = '/media/sf_share/xyz_interval4/'
    # fname = '/media/sf_share/training_set_T{}_interval4'.format(T)
    # fname = '/media/sf_share/T150_training_1000.xyz'
    tset.read(fname)
    configs = tset.get_configs()

    # prediction
    KIMobjs = []
    for i in range(len(configs)):
        obj = KIMcalculator(modelname, params, configs[i])
        obj.initialize()
        KIMobjs.append(obj)

    print('Hello there, started computing Fisher information matrix. It may take a')
    print('while, so have a cup of coffee.')
    fisher = Fisher(KIMobjs, params)
    Fij, Fij_std = fisher.compute()

    # take a look at the SW potential paper
    kB = 8.61733034e-5
    gamma = 0.02
    Fij = Fij / (2.0 * kB * T * gamma)

    # relative variance of Fij
    param_values = params.get_x0()

    # relative variance
    if relative_variance:
        Fij = np.dot(np.dot(np.diag(param_values), Fij), np.diag(param_values))
        Fij_std = np.dot(np.dot(np.diag(param_values), Fij_std), np.diag(param_values))

    with open('Fij', 'w') as fout:
        for line in Fij:
            for i in line:
                fout.write('{:24.16e} '.format(i))
            fout.write('\n')

    with open('Fij_std', 'w') as fout:
        for line in Fij_std:
            for i in line:
                fout.write('{:24.16e} '.format(i))
            fout.write('\n')

    Fij_diag = np.diag(Fij)
    with open('Fij_diag', 'w') as fout:
        for line in Fij_diag:
            fout.write('{:13.5e}\n'.format(line))

    # inverse
    Fij_inv = np.linalg.inv(Fij)
    with open('Fij_inv', 'w') as fout:
        for line in Fij_inv:
            for i in line:
                fout.write('{:24.16e} '.format(i))
            fout.write('\n')

    # inverse_diag
    Fij_inv_diag = np.diag(Fij_inv)
    with open('Fij_inv_diag', 'w') as fout:
        for line in Fij_inv_diag:
            fout.write('{:13.5e}\n'.format(line))

    # eiven analysis
    w, v = np.linalg.eig(Fij)
    with open('eigenVec', 'w') as fout:
        for row in v:
            for item in row:
                fout.write('{:13.5e}'.format(float(item)))
            fout.write('\n')

    with open('eigenVal', 'w') as fout:
        for item in w:
            fout.write('{:13.5e}\n'.format(float(item)))


if __name__ == '__main__':
    # test_fisher(relative_variance=True)
    test_fisher(relative_variance=False)



from collections.abc import Iterable


def length_equal(a, b):
    if isinstance(a, Iterable) and isinstance(b, Iterable):
        if len(a) == len(b):
            return True
        else:
            return False
    # if one is Iterable and the other is not, we treat them as equal since it
    # can be broadcasted
    else:
        return True



import numpy as np
from kliff.dataset import Configuration
from kliff.descriptors.descriptor import Descriptor
from kliff.descriptors.descriptor import load_fingerprints
from kliff.descriptors.descriptor import DescriptorError

# make up some data
num_atoms = 4
num_desc = 5
dim = 3
_zeta = np.arange(num_atoms * num_desc).reshape(num_atoms, num_desc)
_dzeta_dr = np.arange(num_atoms * num_desc * num_atoms * dim)
_dzeta_dr = _dzeta_dr.reshape(num_atoms, num_desc, num_atoms * dim)
_mean = np.mean(_zeta, axis=0)
_stdev = np.std(_zeta, axis=0)
_normalized_zeta = (_zeta - _mean) / _stdev
_normalized_dzeta_dr = _dzeta_dr / np.atleast_3d(_stdev)


def assert_mean_stdev(mean, stdev, target_mean, target_stdev):
    if mean is None:
        assert target_mean is None
    else:
        assert np.allclose(mean, target_mean)
    if stdev is None:
        assert target_stdev is None
    else:
        assert np.allclose(stdev, target_stdev)


def assert_zeta_dzeta_dr(zeta, dzeta_dr, target_zeta, target_dzeta_dr):
    assert np.allclose(zeta, target_zeta)
    if dzeta_dr is None:
        assert target_dzeta_dr is None
    else:
        assert np.allclose(dzeta_dr, target_dzeta_dr)


class ExampleDescriptor(Descriptor):
    def __init__(self, normalize):
        cutvalues = None
        cutname = None
        hyperparams = None
        super(ExampleDescriptor, self).__init__(
            cutvalues, cutname, hyperparams, normalize
        )

    def transform(self, conf, grad):
        if grad:
            return _zeta, _dzeta_dr
        else:
            return _zeta, None


def test_descriptor():
    fname = '../configs_extxyz/Si.xyz'
    conf = Configuration(format='extxyz', identifier=fname)
    conf.read(fname)
    configs = [conf, conf]

    # case 1
    desc = ExampleDescriptor(normalize=False)
    grad = False
    # train set
    desc.generate_train_fingerprints(configs, grad=grad)
    data = load_fingerprints('fingerprints/train.pkl')[0]
    assert_mean_stdev(desc.mean, desc.stdev, None, None)
    assert_zeta_dzeta_dr(data['zeta'], None, _zeta, None)
    # test set
    desc.generate_test_fingerprints(configs, grad=grad)
    data = load_fingerprints('fingerprints/test.pkl')[0]
    assert_mean_stdev(desc.mean, desc.stdev, None, None)
    assert_zeta_dzeta_dr(data['zeta'], None, _zeta, None)

    # case 2
    desc = ExampleDescriptor(normalize=False)
    grad = True
    # train set
    desc.generate_train_fingerprints(configs, grad=grad)
    data = load_fingerprints('fingerprints/train.pkl')[0]
    assert_mean_stdev(desc.mean, desc.stdev, None, None)
    assert_zeta_dzeta_dr(data['zeta'], data['dzeta_dr'], _zeta, _dzeta_dr)
    # test set
    desc.generate_test_fingerprints(configs, grad=grad)
    data = load_fingerprints('fingerprints/test.pkl')[0]
    assert_mean_stdev(desc.mean, desc.stdev, None, None)
    assert_zeta_dzeta_dr(data['zeta'], data['dzeta_dr'], _zeta, _dzeta_dr)

    # case 3
    desc = ExampleDescriptor(normalize=True)
    grad = False
    # train set
    desc.generate_train_fingerprints(configs, grad=grad)
    data = load_fingerprints('fingerprints/train.pkl')[0]
    assert_mean_stdev(desc.mean, desc.stdev, _mean, _stdev)
    assert_zeta_dzeta_dr(data['zeta'], None, _normalized_zeta, None)
    # test set
    desc.generate_test_fingerprints(configs, grad=grad)
    data = load_fingerprints('fingerprints/test.pkl')[0]
    assert_mean_stdev(desc.mean, desc.stdev, _mean, _stdev)
    assert_zeta_dzeta_dr(data['zeta'], None, _normalized_zeta, None)

    # case 4
    desc = ExampleDescriptor(normalize=True)
    grad = True
    # train set
    desc.generate_train_fingerprints(configs, grad=grad)
    data = load_fingerprints('fingerprints/train.pkl')[0]
    assert_mean_stdev(desc.mean, desc.stdev, _mean, _stdev)
    assert_zeta_dzeta_dr(
        data['zeta'], data['dzeta_dr'], _normalized_zeta, _normalized_dzeta_dr
    )
    # test set
    desc.generate_test_fingerprints(configs, grad=grad)
    data = load_fingerprints('fingerprints/test.pkl')[0]
    assert_mean_stdev(desc.mean, desc.stdev, _mean, _stdev)
    assert_zeta_dzeta_dr(
        data['zeta'], data['dzeta_dr'], _normalized_zeta, _normalized_dzeta_dr
    )

    # case 5 allow reuse
    desc = ExampleDescriptor(normalize=True)
    grad = True
    # train set
    desc.generate_train_fingerprints(configs, grad=grad, reuse=True)
    data = load_fingerprints('fingerprints/train.pkl')[0]
    assert_mean_stdev(desc.mean, desc.stdev, _mean, _stdev)
    assert_zeta_dzeta_dr(
        data['zeta'], data['dzeta_dr'], _normalized_zeta, _normalized_dzeta_dr
    )
    # test set
    desc.generate_test_fingerprints(configs, grad=grad, reuse=True)
    data = load_fingerprints('fingerprints/test.pkl')[0]
    assert_mean_stdev(desc.mean, desc.stdev, _mean, _stdev)
    assert_zeta_dzeta_dr(
        data['zeta'], data['dzeta_dr'], _normalized_zeta, _normalized_dzeta_dr
    )

    # try use generate_test_fingerprints() before genereate_train_fingerprints()
    # case 1 (should work, since we do not require normalize)
    desc = ExampleDescriptor(normalize=False)
    desc.generate_test_fingerprints(configs, grad=True)
    data = load_fingerprints('fingerprints/test.pkl')[0]
    assert_mean_stdev(desc.mean, desc.stdev, None, None)
    assert_zeta_dzeta_dr(data['zeta'], data['dzeta_dr'], _zeta, _dzeta_dr)

    # case 2 (should not work, since we do need normalize)
    desc = ExampleDescriptor(normalize=True)
    try:
        desc.generate_test_fingerprints(configs, grad=True)
    except DescriptorError as e:
        assert e.__class__ == DescriptorError


if __name__ == '__main__':
    test_descriptor()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
import tensorflow as tf
from openkim_fit.dataset import DataSet
from openkim_fit.descriptor import Descriptor
from tensorflow.contrib.layers import (
    fully_connected,
    xavier_initializer,
    variance_scaling_initializer,
)
import openkim_fit.ann as ann

# Check that the forces and energys are the same with the KIM model is used.
# 0. To test dropout, modify `dropout_` function in `network.cpp` to set `debug` to true,
#    where the dropout is hard-coded to be the same as the one here.
#    To not test dropout, do not need to change it.
# 1. run this script
# 2. copy `ann_kim.params' generated by this script to the KIM Carbon model, and make.
#    Note, if `struct == bilayer`, copy to `Artificial_Neural_Network_C__MO_000000111111_000`
#    elif `struct == bulk`, copy to `Artificial_Neural_Network_Intralayer_C__MO_000000111111_000`.
# 3. run LAMMPS with input `lammps_graphene_bilayer.in` or `lammps_graphene_monolayer.in' to generate
#    the dump file. (located at bop  /usr/scratch/wenxx151/Documents/lmp_playground/debug_ann)
# 4. compare the stdout of this script and the LAMMPS dump file.
# 5. Agreement between models you want to check when this test does not pass.
# 1) Both use the same STRUCT.
# 2) This script will write the not normalized descriptor values to file `debug_descriptor.txt`,
#    and the normalized descriptor values to `gc_normalized.txt`.
#    It can be compared with the one generated by the model driver. TO enable it, search `TODO`
#    in ANNImplementation.hpp.
#    Note, look carefully for the order of the descriptor values for each atom. They may be different
#    from this script and the KIM model. Here, in function `dropout` we are modifying those of the first
#    3 atoms. If the order of the descriptor value of the first threee atoms are differert, different
#    results will be obtained.


# set a global random seed
tf.set_random_seed(1)


##############################
# settings
# DO_NORMALIZE = False
DO_NORMALIZE = True
# DO_DROPOUT = False
DO_DROPOUT = True
STRUCT = 'bulk'
# STRUCT = 'bilayer'
##############################

if DO_DROPOUT == True:
    keep_prob = [0.9, 1, 0.7]
else:
    keep_prob = [1.0, 1.0, 1.0]


def dropout(x, keep_prob):
    if keep_prob < 1 - 1e-10:
        binary = np.ones(x.shape)
        binary[0][0] = 0.0
        binary[0][5] = 0.0
        binary[1][2] = 0.0
        binary[1][7] = 0.0
        binary[2][2] = 0.0
        binary[2][5] = 0.0
        binary = tf.constant(binary)
        return (x / keep_prob) * binary
    else:
        return x


# create Descriptor
cutfunc = 'cos'
cutvalue = {'C-C': 5.0}
# cutvalue = {'Mo-Mo':5., 'Mo-S':5., 'S-S':5.}
desc_params = {
    'g1': None,
    'g2': [{'eta': 0.1, 'Rs': 0.2}, {'eta': 0.3, 'Rs': 0.4}],
    'g3': [{'kappa': 0.1}, {'kappa': 0.2}, {'kappa': 0.3}],
    'g4': [
        {'zeta': 0.1, 'lambda': 0.2, 'eta': 0.01},
        {'zeta': 0.3, 'lambda': 0.4, 'eta': 0.02},
    ],
    'g5': [
        {'zeta': 0.11, 'lambda': 0.22, 'eta': 0.011},
        {'zeta': 0.33, 'lambda': 0.44, 'eta': 0.022},
    ],
}

desc = Descriptor(desc_params, cutfunc, cutvalue, cutvalue_samelayer=cutvalue, debug=True)
num_desc = desc.get_num_descriptors()


# read config and reference data
tset = DataSet()
if STRUCT == 'bilayer':
    tset.read('./training_set/graphene_bilayer_1x1.xyz')
else:
    tset.read('./training_set/graphene_monolayer_2x2.xyz')
configs = tset.get_configs()


# preprocess data to generate tfrecords
DTYPE = tf.float64
train_name, _ = ann.convert_to_tfrecord(
    configs,
    desc,
    size_validation=0,
    directory='/tmp',
    do_generate=True,
    do_normalize=DO_NORMALIZE,
    do_shuffle=False,
    use_welford=False,
    fit_forces=True,
    structure=STRUCT,
    dtype=DTYPE,
)

# read data from tfrecords into tensors
dataset = ann.read_tfrecord(train_name, fit_forces=True, dtype=DTYPE)

# number of epoches
iterator = dataset.make_one_shot_iterator()


# create NN
initializer = ann.weight_decorator(xavier_initializer(dtype=DTYPE))
size = 20
subloss = []

name, num_atoms_by_species, weight, gen_coords, energy_label, atomic_coords, dgen_datomic_coords, forces_label = (
    iterator.get_next()
)

in_layer = ann.input_layer_given_data(
    atomic_coords, gen_coords, dgen_datomic_coords, num_descriptor=num_desc
)
in_layer.set_shape((configs[0].get_num_atoms(), num_desc))
in_layer_drop = dropout(in_layer, keep_prob[0])


hidden1 = fully_connected(
    in_layer_drop,
    size,
    activation_fn=tf.nn.tanh,
    weights_initializer=initializer,
    biases_initializer=tf.truncated_normal_initializer(dtype=DTYPE),
    scope='hidden1',
)
hidden1_drop = dropout(hidden1, keep_prob[1])

hidden2 = fully_connected(
    hidden1_drop,
    size,
    activation_fn=tf.nn.tanh,
    weights_initializer=initializer,
    biases_initializer=tf.truncated_normal_initializer(dtype=DTYPE),
    scope='hidden2',
)
hidden2_drop = dropout(hidden2, keep_prob[2])

output = fully_connected(
    hidden2_drop,
    1,
    activation_fn=None,
    weights_initializer=initializer,
    biases_initializer=tf.truncated_normal_initializer(dtype=DTYPE),
    scope='output',
)


# energy and forces
energy = tf.reduce_sum(output)
# tf.gradients return a LIST of tensors
forces = -tf.gradients(output, atomic_coords)[0]

weights, biases = ann.get_weights_and_biases(['hidden1', 'hidden2', 'output'])


with tf.Session() as sess:

    # init global vars
    init_op = tf.global_variables_initializer()
    sess.run(init_op)

    en, fo = sess.run([energy, forces])
    print('energy:', en)
    print('forces:')
    for i, f in enumerate(fo):
        print('{:13.5e}'.format(f), end='')
        if i % 3 == 2:
            print()

    # output results to a KIM model
    w, b = sess.run([weights, biases])
    ann.write_kim_ann(desc, w, b, tf.nn.tanh, keep_prob=keep_prob, dtype=tf.float64)


# write normalized gcc
tfrecord_name = '/tmp/train.tfrecord'
text_name = 'gc_normalized.txt'
ann.tfrecord_to_text(tfrecord_name, text_name, fit_forces=True, dtype=tf.float64)



import os
import numpy as np
import scipy.optimize
import multiprocessing as mp
import kliff
from kliff import parallel
from kliff.error import InputError

try:
    import torch
    from kliff.neuralnetwork import FingerprintsDataset
    from kliff.neuralnetwork import FingerprintsDataLoader

    torch_available = True
except ImportError:
    torch_available = False

logger = kliff.logger.get_logger(__name__)


def energy_forces_residual(identifier, natoms, prediction, reference, data):
    """A residual function using both energy and forces.

    Parameters
    ----------
    identifier: str
        identifer of the configuration, i.e. path to the file

    natoms: int
        number of atoms in the configuration

    prediction: 1D array
        prediction computed by calculator

    reference: 1D array
        references data for the prediction

    data: dict
        User provided callback data, and aviailable keys:
            energy_weight: float (default: 1)
            forces_weight: float (default: 1)

    normalize: bool (default: False)
        Whether to normalize the residual by the number of atoms

    Return
    ------
    residual: 1D array

    Note
    ----
    The length of `prediction` and `reference` (call it `S`) are the same, and it
    depends on `use_energy` and `use_forces` in KIMCalculator. Assume the
    configuration contains of `N` atoms.

    1) If `use_energy == True` and `use_forces == False`, then `S = 1`.
    `prediction[0]` is the potential energy computed by the calculator, and
    `reference[0]` is the reference energy.

    2) If `use_energy == False` and `use_forces == True`, then `S = 3N`.
    `prediction[3*i+0]`, `prediction[3*i+1]`, and `prediction[3*i+2]` are the
    x, y, and z component of the forces on atom i in the configuration, respecrively.
    Correspondingly, `reference` is the 3N concatenated reference forces.


    3) If `use_energy == True` and `use_forces == True`, then `S = 3N + 1`.
    `prediction[0]` is the potential energy computed by the calculator, and
    `reference[0]` is the reference energy.
    `prediction[3*i+1]`, `prediction[3*i+2]`, and `prediction[3*i+3]` are the
    x, y, and z component of the forces on atom i in the configuration, respecrively.
    Correspondingly, `reference` is the 3N concatenated reference forces.
    """

    if len(prediction) != 3 * natoms + 1:
        raise ValueError("len(prediction) != 3N+1, where N is the number of atoms.")

    # prepare weight based on user provided data
    energy_weight = data['energy_weight']
    forces_weight = data['forces_weight']
    normalize = data['normalize_by_natoms']
    if energy_weight is None:
        energy_weight = 1.0
    if forces_weight is None:
        forces_weight = 1.0
    if normalize:
        energy_weight /= natoms
        forces_weight /= natoms

    # obtain residual and properly normalize it
    residual = prediction - reference
    residual[0] *= energy_weight
    residual[1:] *= forces_weight

    return residual


def forces_residual(conf_id, natoms, prediction, reference, data):
    if data is not None:
        data['energy_weight'] = 0
    else:
        data = {'energy_weight': 0}
    return energy_forces_residual(conf_id, natoms, prediction, reference, data)


def energy_residual(conf_id, natoms, prediction, reference, data):
    if data is not None:
        data['forces_weight'] = 0
    else:
        data = {'forces_weight': 0}
    return energy_forces_residual(conf_id, natoms, prediction, reference, data)


class Loss(object):
    """Objective function class to conduct the optimization."""

    def __new__(
        self, calculator, nprocs=1, residual_fn=energy_forces_residual, residual_data=None
    ):
        """
        Parameters
        ----------

        calculator: Calculator object

        nprocs: int
          Number of processors to parallel to run the predictors.

        residual_fn: function
          function to compute residual, see `energy_forces_residual` for an example

        residual_data: dict
          data passed to residual function

        """
        data = self.check_residual_data(residual_data)

        calc_type = calculator.__class__.__name__

        if calc_type == 'PytorchANNCalculator':
            return LossNeuralNetworkModel(calculator, nprocs, residual_fn, data)
        else:
            return LossPhysicsMotivatedModel(calculator, nprocs, residual_fn, data)

    @staticmethod
    def check_residual_data(data):
        default = {
            'energy_weight': 1.0,
            'forces_weight': 1.0,
            'stress_weight': 1.0,
            'normalize_by_natoms': True,
        }
        if data is not None:
            for key, value in data.items():
                if key not in default:
                    raise InputError('"{}" not supported by "residual_data".'.format(key))
                else:
                    default[key] = value
        return default


class LossPhysicsMotivatedModel(object):
    """Objective function class to conduct the optimization for PM models."""

    scipy_minimize_methods = [
        'Nelder-Mead',
        'Powell',
        'CG',
        'BFGS',
        'Newton-CG',
        'L-BFGS-B',
        'TNC',
        'COBYLA',
        'SLSQP',
        'trust-constr',
        'dogleg',
        'trust-ncg',
        'trust-exact',
        'trust-krylov',
    ]

    scipy_minimize_methods_not_supported_arguments = ['bounds']

    scipy_least_squares_methods = ['trf', 'dogbox', 'lm']

    scipy_least_squares_methods_not_supported_arguments = ['bounds']

    def __init__(
        self, calculator, nprocs=1, residual_fn=energy_forces_residual, residual_data=None
    ):
        """
        Parameters
        ----------

        calculator: Calculator object

        nprocs: int
          Number of processors to parallel to run the predictors.

        residual_fn: function
          function to compute residual, see `energy_forces_residual` for an example

        residual_data: dict
          data passed to residual function

        """

        self.calculator = calculator
        self.nprocs = nprocs

        self.residual_fn = residual_fn
        self.residual_data = residual_data if residual_data is not None else dict()
        self.calculator_type = calculator.__class__.__name__

        if self.calculator_type == 'WrapperCalculator':
            calculators = self.calculator.calculators
        else:
            calculators = [self.calculator]
        for calc in calculators:
            infl_dist = calc.model.get_influence_distance()
            cas = calc.get_compute_arguments()
            # TODO can be parallelized
            for ca in cas:
                ca.refresh(infl_dist)

        logger.info('"{}" instantiated.'.format(self.__class__.__name__))

    #
    #  def set_nprocs(self, nprocs):
    #    """ Set the number of processors to be used."""
    #    self.nprocs = nprocs
    #
    #  def set_residual_fn_and_data(self, fn, data):
    #    """ Set residual function and data. """
    #    self.residual_fn = fn
    #    self.residual_fn_data = data

    def minimize(self, method, **kwargs):
        """ Minimize the loss.

        Parameters
        ----------
        method: str
            minimization methods as specified at:
            https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
            https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html
        kwargs: extra keyword arguments that can be used by the scipy optimizer.
        """

        if method in self.scipy_least_squares_methods:
            for i in self.scipy_least_squares_methods_not_supported_arguments:
                if i in kwargs:
                    raise LossError(
                        'Argument "{}" should not be set through the '
                        '"minimize" method.'.format(i)
                    )
            bounds = self.calculator.get_opt_params_bounds()
            kwargs['bounds'] = bounds
            logger.info(
                'Start minimization using scipy least squares method: {}.'.format(method)
            )
            result = self._scipy_optimize_least_squares(method, **kwargs)
            logger.info(
                'Finish minimization using scipy least squares method: {}.'.format(method)
            )

        elif method in self.scipy_minimize_methods:
            for i in self.scipy_minimize_methods_not_supported_arguments:
                if i in kwargs:
                    raise LossError(
                        'Argument "{}" should not be set through the '
                        '"minimize" method.'.format(i)
                    )
            bounds = self.calculator.get_opt_params_bounds()
            for i in range(len(bounds)):
                lb = bounds[i][0]
                ub = bounds[i][1]
                if lb is None:
                    bounds[i][0] = -np.inf
                if ub is None:
                    bounds[i][1] = np.inf
            kwargs['bounds'] = bounds
            logger.info(
                'Start minimization using scipy optimize method: {}.'.format(method)
            )
            result = self._scipy_optimize_minimize(method, **kwargs)
            logger.info(
                'Finish minimization using scipy optimize method: {}.'.format(method)
            )

        else:
            raise LossError('minimization method "{}" not supported.'.format(method))

        # update final optimized paramters
        self.calculator.update_opt_params(result.x)
        return result

    def get_residual(self, x):
        """ Compute the residual.

        This is a callable for optimizing method in scipy.optimize.least_squares,
        which is passed as the first positional argument.

        Parameters
        ----------

        x: 1D array
          optimizing parameter values
        """

        # publish params x to predictor
        self.calculator.update_opt_params(x)

        cas = self.calculator.get_compute_arguments()

        if self.calculator_type == 'WrapperCalculator':
            calc_list = self.calculator.get_calculator_list()
            X = zip(cas, calc_list)
            if self.nprocs > 1:
                residuals = parallel.parmap2(
                    self._get_residual_single_config,
                    X,
                    self.residual_fn,
                    self.residual_data,
                    nprocs=self.nprocs,
                    tuple_X=True,
                )
                residual = np.concatenate(residuals)
            else:
                residual = []
                for ca, calc in X:
                    current_residual = self._get_residual_single_config(
                        ca, calc, self.residual_fn, self.residual_data
                    )
                    residual = np.concatenate((residual, current_residual))

        else:
            if self.nprocs > 1:
                residuals = parallel.parmap2(
                    self._get_residual_single_config,
                    cas,
                    self.calculator,
                    self.residual_fn,
                    self.residual_data,
                    nprocs=self.nprocs,
                    tuple_X=False,
                )
                residual = np.concatenate(residuals)
            else:
                residual = []
                for ca in cas:
                    current_residual = self._get_residual_single_config(
                        ca, self.calculator, self.residual_fn, self.residual_data
                    )
                    residual = np.concatenate((residual, current_residual))

        return residual

    def get_loss(self, x):
        """Compute the loss.

        This is a callable for optimizing method in scipy.optimize,minimize,
        which is passed as the first positional argument.

        Parameters
        ----------

        x: 1D array
          optimizing parameter values
        """
        residual = self.get_residual(x)
        loss = 0.5 * np.linalg.norm(residual) ** 2
        return loss

    def _scipy_optimize_least_squares(self, method, **kwargs):
        residual = self.get_residual
        x = self.calculator.get_opt_params()
        return scipy.optimize.least_squares(residual, x, method=method, **kwargs)

    def _scipy_optimize_minimize(self, method, **kwargs):
        loss = self.get_loss
        x = self.calculator.get_opt_params()
        return scipy.optimize.minimize(loss, x, method=method, **kwargs)

    def _get_residual_single_config(self, ca, calculator, residual_fn, residual_data):

        # prediction data
        calculator.compute(ca)
        pred = calculator.get_prediction(ca)

        # reference data
        ref = calculator.get_reference(ca)

        conf = ca.conf
        identifier = conf.get_identifier()
        natoms = conf.get_number_of_atoms()

        residual = residual_fn(identifier, natoms, pred, ref, residual_data)

        return residual

    def __enter__(self):
        return self

    def __exit__(self, exec_type, exec_value, trackback):
        # if there is expections, raise it (not for KeyboardInterrupt)
        if exec_type is not None and exec_type is not KeyboardInterrupt:
            return False  # return False will cause Python to re-raise the expection


class LossNeuralNetworkModel(object):
    """Objective function class to conduct the optimization for ML models."""

    torch_minimize_methods = [
        'Adadelta',
        'Adagrad',
        'Adam',
        'SparseAdam',
        'Adamax',
        'ASGD',
        'LBFGS',
        'RMSprop',
        'Rprop',
        'SGD',
    ]

    def __init__(
        self, calculator, nprocs=1, residual_fn=energy_forces_residual, residual_data=None
    ):
        """
        Parameters
        ----------

        calculator: Calculator object

        nprocs: int
          Number of processors to parallel to run the predictors.

        residual_fn: function
          Function to compute residual, see `energy_forces_residual` for an example.

        residual_data: dict
          Data passed to residual function.

        """

        if not torch_available:
            raise ImportError('Please install "PyTorch" first. See: https://pytorch.org')

        self.calculator = calculator
        self.nprocs = nprocs

        self.residual_fn = residual_fn
        self.residual_data = residual_data if residual_data is not None else dict()

        self.calculator_type = calculator.__class__.__name__

        logger.info('"{}" instantiated.'.format(self.__class__.__name__))

    #
    #  def set_nprocs(self, nprocs):
    #    """ Set the number of processors to be used."""
    #    self.nprocs = nprocs
    #
    #  def set_residual_fn_and_data(self, fn, data):
    #    """ Set residual function and data. """
    #    self.residual_fn = fn
    #    self.residual_fn_data = data

    def minimize(self, method, batch_size=100, num_epochs=1000, **kwargs):
        """ Minimize the loss.

        Parameters
        ----------
        method: str
            PyTorch optimization methods, and aviliable ones are:
            [`Adadelta`, `Adagrad`, `Adam`, `SparseAdam`, `Adamax`, `ASGD`,
            `LBFGS`, `RMSprop`, `Rprop`, `SGD`]
            See also: https://pytorch.org/docs/stable/optim.html

        kwargs: extra keyword arguments that can be used by the PyTorch optimizer.
        """
        if method not in self.torch_minimize_methods:
            raise LossError('Minimization method "{}" not supported.'.format(method))
        self.method = method
        self.batch_size = batch_size
        self.num_epochs = num_epochs

        # data loader
        fname = self.calculator.get_train_fingerprints_path()
        fp = FingerprintsDataset(fname)
        self.data_loader = FingerprintsDataLoader(dataset=fp, num_epochs=self.num_epochs)

        # optimizing
        try:
            optimizer = getattr(torch.optim, method)(
                self.calculator.model.parameters(), **kwargs
            )
        except TypeError as e:
            print(str(e))
            idx = str(e).index("argument '") + 10
            err_arg = str(e)[idx:].strip("'")
            raise InputError(
                'Argument "{}" not supported by optimizer "{}".'.format(err_arg, method)
            )

        # model save metadata
        save_prefix = self.calculator.model.save_prefix
        save_start = self.calculator.model.save_start
        save_frequency = self.calculator.model.save_frequency
        if save_prefix is None or save_start is None or save_frequency is None:
            logger.info(
                'Model saving meta data not set by user. Now set it '
                'to "prefix=./kliff_saved_model", "start=num_epochs/2",'
                ' and "frequency=10".'
            )
            save_prefix = os.path.join(os.getcwd(), 'kliff_saved_model')
            save_start = self.num_epochs // 2
            save_frequency = 10
            self.calculator.model.set_save_metadata(
                save_prefix, save_start, save_frequency
            )

        # other metadata
        n = 0
        epoch = 0
        DATASET_SIZE = len(self.calculator.configs)

        msg = 'Start minimization using optimization method: {}.'.format(self.method)
        logger.info(msg)
        print(msg)

        while True:
            try:
                if self.method in ['LBFGS']:

                    def closure():
                        # zero the parameter gradients
                        optimizer.zero_grad()
                        # forward + backward + optimize
                        loss = self.get_loss()
                        loss.backward()
                        return loss

                    optimizer.step(closure)
                else:
                    optimizer.zero_grad()
                    # forward + backward + optimize
                    loss = self.get_loss()
                    loss.backward()
                    optimizer.step()
                epoch_new = n * self.batch_size // DATASET_SIZE

                if epoch_new > epoch:
                    epoch = epoch_new
                    print('Epoch = {}, loss = {}'.format(epoch, loss))
                n += 1

                if (
                    epoch_new >= save_start
                    and (epoch_new - save_start) % save_frequency == 0
                ):
                    fname = 'model_epoch{}.pkl'.format(epoch)
                    path = os.path.join(save_prefix, fname)
                    self.calculator.model.save(path)

            except StopIteration:
                break

    def get_loss(self):
        loss = 0
        for _ in range(self.batch_size):
            # raise StopIteration error if out of bounds; This will ignore the last
            # chunk of data whose size is smaller than `batch_size`
            inp = self.data_loader.next_element()
            residual = self._get_residual_single_config(
                inp, self.calculator, self.residual_fn, self.residual_data
            )
            c = torch.sum(torch.pow(residual, 2))
            loss += c
        # TODO  maybe divide batch_size elsewhere
        loss /= self.batch_size
        return loss

    def _get_residual_single_config(self, inp, calculator, residual_fn, residual_data):

        # prediction data
        results = calculator.compute(inp)
        pred_energy = results['energy']
        pred_forces = results['forces']

        if calculator.use_energy:
            pred = torch.tensor([pred_energy])
            ref = torch.tensor([inp['energy'][0]])
        if calculator.use_forces:
            ref_forces = inp['forces'][0]
            if calculator.use_energy:
                pred = torch.cat((pred, pred_forces.reshape((-1,))))
                ref = torch.cat((ref, ref_forces.reshape((-1,))))
            else:
                pred = pred_forces.reshape((-1,))
                ref = ref_forces.reshape((-1,))

        identifier = inp['name'][0]
        species = inp['species'][0]
        natoms = len(species)

        residual = residual_fn(identifier, natoms, pred, ref, residual_data)

        return residual


class LossError(Exception):
    def __init__(self, msg):
        super(LossError, self).__init__(msg)
        self.msg = msg

    def __expr__(self):
        return self.msg



# Training a model using multiple processes:
import torch
import torch.nn as nn
import torch.multiprocessing as mp
import numpy as np
from torch.utils.data import TensorDataset, DataLoader

torch.manual_seed(35)


X = torch.empty(20, 5).uniform_(0, 1)
Y = np.asarray(np.atleast_2d(0.1 * np.arange(20)).T, dtype=np.float32)
Y = torch.from_numpy(Y)
dataset = TensorDataset(X, Y)
dataset = DataLoader(dataset, batch_size=4, shuffle=True, num_workers=0)


def loss_fn(pref, label):
    return torch.sum(torch.pow(pref - label, 2))


def train(model, tset):
    optimizer = torch.optim.Adam(model.parameters(), lr=0.01)
    for data, labels in tset:
        print('@@lable', labels)
        optimizer.zero_grad()
        loss_fn(model(data), labels).backward()
        optimizer.step()  # This will update the shared parameters


n_in = 5
n_h1 = 10
n_out = 1
model = nn.Sequential(nn.Linear(n_in, n_h1), nn.ReLU(), nn.Linear(n_h1, n_out))
model.share_memory()  # Required for 'fork' method to work

print('@@ params before:', model.parameters())
for i in model.parameters():
    print(i)
    break

processes = []
# for i in range(mp.cpu_count()):  # No. of processes
for i in range(1):  # No. of processes
    print('@@@ cpu i', i)
    p = mp.Process(target=train, args=(model, dataset))
    p.start()
    processes.append(p)
for p in processes:
    p.join()
print('@@ params after:', model.parameters())
for i in model.parameters():
    print(i)
    break



import numpy as np
import sys
import logging
from collections import OrderedDict
import kliff
from kliff.descriptors.descriptor import Descriptor
from kliff.descriptors.descriptor import generate_full_cutoff, generate_species_code
from kliff.neighbor import NeighborList
from . import sf

logger = kliff.logger.get_logger(__name__)


class SymmetryFunction(Descriptor):
    """Atom-centered symmetry functions descriptor as discussed in [Behler2011]_.

    Parameters
    ----------
    cut_dists: dict
        Cutoff distances, with key of the form ``A-B`` where ``A`` and ``B`` are
        atomic species string, and value should be a float.

    cut_name: str
        Name of the cutoff function.

    hyperparams: dict or str
        A dictionary of the hyper parameters of that define the descriptor. We
        provide two sets of hyperparams that can be used by setting
        ``hyperparams='set51'`` or ``hyperparams='set31'``, which are taken from
        [Artrith2012]_ and [Artrith2013]_, respectively. To see what they are, one
        can do:

        >>> cut_name = 'cos'  # just for init purpose
        >>> cut_dists = {'C-C': 5.}  # just for init purpose
        >>> hyperparams = 'set51'
        >>> desc = SymmetryFunction(cut_dists, cut_name, hyperparams)
        >>> desc.get_hyperparams()

    normalize: bool (optional)
        If ``True``, the fingerprints is centered and normalized according to:
        ``zeta = (zeta - mean(zeta)) / stdev(zeta)``

    dtype: np.dtype (optional)
        Data type for the generated fingerprints, such as ``np.float32`` and
        ``np.float64``.

    Example
    -------

    If ``set51`` or ``set31`` hyperparams are used, the cutoff distances should be
    given in ``Angstrom``.

    >>> cut_name = 'cos'
    >>> cut_dists = {'C-C': 5., 'C-H': 4.5, 'H-H': 4.0}
    >>> hyperparams = 'set51'
    >>> desc = SymmetryFunction(cut_dists, cut_name, hyperparams)

    You can provide your own hyperparams as a dictionary:

    >>> cut_name = 'cos'
    >>> cut_dists = {'C-C': 5., 'C-H': 4.5, 'H-H': 4.0}
    >>> hyperparams = {'g1': None,
    >>>                'g2': [{'eta':0.1, 'Rs':0.2}, {'eta':0.3, 'Rs':0.4}],
    >>>                'g3': [{'kappa':0.1}, {'kappa':0.2}, {'kappa':0.3}]}
    >>> desc = SymmetryFunction(cut_dists, cut_name, hyperparams)

    References
    ----------
    .. [Behler2011] J. Behler, "Atom-centered symmetry functions for constructing
       high-dimensional neural network potentials," J. Chem. Phys. 134, 074106
       (2011).
    .. [Artrith2012] N. Artrith and J. Behler. "High-dimensional neural network
       potentials for metal surfaces: A prototype study for copper." Physical Review
       B 85, no. 4 (2012): 045439.
    .. [Artrith2013] N. Artrith, B. Hiller, and J. Behler. "Neural network potentials
       for metals and oxides–First applications to copper clusters at zinc oxide."
       physica status solidi (b) 250, no. 6 (2013): 1191-1203.
    """

    def __init__(
        self, cut_dists, cut_name, hyperparams, normalize=True, dtype=np.float32
    ):
        super(SymmetryFunction, self).__init__(
            cut_dists, cut_name, hyperparams, normalize, dtype
        )

        self._desc = OrderedDict()

        self._cdesc = sf.Descriptor()
        self._set_cutoff()
        self._set_hyperparams()

        logger.info('"SymmetryFunction" descriptor initialized.')

    def transform(self, conf, grad=False):
        """Transform atomic coords to atomic enviroment descriptor values.

        Parameters
        ----------
        conf: :class:`~kliff.dataset.Configuration` object
            A configuration of atoms.


        grad: bool (optional)
            Whether to compute the gradient of descriptor values w.r.t. atomic
            coordinates.

        Returns
        -------
        zeta: 2D array
            Descriptor values, each row for one atom.
            zeta has shape (num_atoms, num_descriptors), where num_atoms is the
            number of atoms in the configuration, and num_descriptors is the size
            of the descriptor vector (depending on the the choice of hyper-parameters).

        dzeta_dr: 4D array if grad is ``True``, otherwise ``None``
            Gradient of descriptor values w.r.t. atomic coordinates.
            dzeta_dr has shape (num_atoms, num_descriptors, num_atoms, DIM), where
            num_atoms and num_descriptors has the same meanings as described in zeta.
            DIM = 3 denotes three Cartesian coordinates.
        """

        # create neighbor list
        infl_dist = max(self.cutoff.values())
        nei = NeighborList(conf, infl_dist, padding_need_neigh=False)

        coords = nei.coords
        image = nei.image
        species = np.asarray([self.species_code[i] for i in nei.species], dtype=np.intc)
        numneigh, neighlist = nei.get_numneigh_and_neighlist_1D()

        Natoms = len(coords)
        Ncontrib = conf.get_number_of_atoms()
        Ndesc = len(self)

        if grad:
            zeta, dzeta_dr = self._cdesc.get_gen_coords_and_deri(
                coords, species, neighlist, numneigh, image, Natoms, Ncontrib, Ndesc
            )
            # reshape 3D array to 4D array
            dzeta_dr = dzeta_dr.reshape(Ncontrib, Ndesc, Ncontrib, 3)
        else:
            zeta = self._cdesc.get_gen_coords(
                coords, species, neighlist, numneigh, image, Natoms, Ncontrib, Ndesc
            )
            dzeta_dr = None

        if logger.getEffectiveLevel() == logging.DEBUG:
            logger.debug(
                '\n' + '=' * 25 + 'descriptor values (no normalization)' + '=' * 25
            )
            logger.debug('\nconfiguration name: %s', conf.get_identifier())
            logger.debug('\natom id    descriptor values ...')
            for i, line in enumerate(zeta):
                s = '\n{}    '.format(i)
                for j in line:
                    s += '{:.15g} '.format(j)
                logger.debug(s)

        return zeta, dzeta_dr

    def _set_cutoff(self):
        supported = ['cos', 'exp']
        if self.cut_name is None:
            self.cut_name = supported[0]
        if self.cut_name not in supported:
            spd = ['"{}", '.format(s) for s in supported]
            raise SymmetryFunctionError(
                'Cutoff "{}" not supported by this descriptor. Use {}.'.format(
                    self.cut_name, spd
                )
            )

        self.cutoff = generate_full_cutoff(self.cut_dists)
        self.species_code = generate_species_code(self.cut_dists)
        num_species = len(self.species_code)

        rcutsym = np.zeros([num_species, num_species], dtype=np.double)
        for si, i in self.species_code.items():
            for sj, j in self.species_code.items():
                rcutsym[i][j] = self.cutoff[si + '-' + sj]
        self._cdesc.set_cutoff(self.cut_name, rcutsym)

    def _set_hyperparams(self):
        if isinstance(self.hyperparams, str):
            name = self.hyperparams.lower()
            if name == 'set51':
                self.hyperparams = get_set51()
            elif name == 'set31':
                self.hyperparams = get_set31()
            else:
                raise SymmetryFunctionError('hyperparams "{}" unrecognized.'.format(name))
        if not isinstance(self.hyperparams, OrderedDict):
            self.hyperparams = OrderedDict(self.hyperparams)

        # hyperparams of descriptors
        for key, values in self.hyperparams.items():
            if key.lower() not in ['g1', 'g2', 'g3', 'g4', 'g5']:
                raise SymmetryFunctionError(
                    'Symmetry function "{}" unrognized.'.format(key)
                )

            # g1 needs no hyperparams, put a placeholder
            name = key.lower()
            if name == 'g1':
                # it has no hyperparams, zeros([1,1]) for placeholder
                params = np.zeros([1, 1], dtype=np.double)
            else:
                rows = len(values)
                cols = len(values[0])
                params = np.zeros([rows, cols], dtype=np.double)
                for i, line in enumerate(values):
                    if name == 'g2':
                        params[i][0] = line['eta']
                        params[i][1] = line['Rs']
                    elif name == 'g3':
                        params[i][0] = line['kappa']
                    elif key == 'g4':
                        params[i][0] = line['zeta']
                        params[i][1] = line['lambda']
                        params[i][2] = line['eta']
                    elif key == 'g5':
                        params[i][0] = line['zeta']
                        params[i][1] = line['lambda']
                        params[i][2] = line['eta']

            # store cutoff values in both this python and cpp class
            self._desc[name] = params
            self._cdesc.add_descriptor(name, params)

    def __len__(self):
        N = 0
        for key in self._desc:
            N += len(self._desc[key])
        return N

    def get_size(self):
        return len(self)

    def get_hyperparams(self):
        return self._desc


def get_set51():

    params = OrderedDict()

    params['g2'] = [
        {'eta': 0.001, 'Rs': 0.0},
        {'eta': 0.01, 'Rs': 0.0},
        {'eta': 0.02, 'Rs': 0.0},
        {'eta': 0.035, 'Rs': 0.0},
        {'eta': 0.06, 'Rs': 0.0},
        {'eta': 0.1, 'Rs': 0.0},
        {'eta': 0.2, 'Rs': 0.0},
        {'eta': 0.4, 'Rs': 0.0},
    ]

    params['g4'] = [
        {'zeta': 1, 'lambda': -1, 'eta': 0.0001},
        {'zeta': 1, 'lambda': 1, 'eta': 0.0001},
        {'zeta': 2, 'lambda': -1, 'eta': 0.0001},
        {'zeta': 2, 'lambda': 1, 'eta': 0.0001},
        {'zeta': 1, 'lambda': -1, 'eta': 0.003},
        {'zeta': 1, 'lambda': 1, 'eta': 0.003},
        {'zeta': 2, 'lambda': -1, 'eta': 0.003},
        {'zeta': 2, 'lambda': 1, 'eta': 0.003},
        {'zeta': 1, 'lambda': -1, 'eta': 0.008},
        {'zeta': 1, 'lambda': 1, 'eta': 0.008},
        {'zeta': 2, 'lambda': -1, 'eta': 0.008},
        {'zeta': 2, 'lambda': 1, 'eta': 0.008},
        {'zeta': 1, 'lambda': -1, 'eta': 0.015},
        {'zeta': 1, 'lambda': 1, 'eta': 0.015},
        {'zeta': 2, 'lambda': -1, 'eta': 0.015},
        {'zeta': 2, 'lambda': 1, 'eta': 0.015},
        {'zeta': 4, 'lambda': -1, 'eta': 0.015},
        {'zeta': 4, 'lambda': 1, 'eta': 0.015},
        {'zeta': 16, 'lambda': -1, 'eta': 0.015},
        {'zeta': 16, 'lambda': 1, 'eta': 0.015},
        {'zeta': 1, 'lambda': -1, 'eta': 0.025},
        {'zeta': 1, 'lambda': 1, 'eta': 0.025},
        {'zeta': 2, 'lambda': -1, 'eta': 0.025},
        {'zeta': 2, 'lambda': 1, 'eta': 0.025},
        {'zeta': 4, 'lambda': -1, 'eta': 0.025},
        {'zeta': 4, 'lambda': 1, 'eta': 0.025},
        {'zeta': 16, 'lambda': -1, 'eta': 0.025},
        {'zeta': 16, 'lambda': 1, 'eta': 0.025},
        {'zeta': 1, 'lambda': -1, 'eta': 0.045},
        {'zeta': 1, 'lambda': 1, 'eta': 0.045},
        {'zeta': 2, 'lambda': -1, 'eta': 0.045},
        {'zeta': 2, 'lambda': 1, 'eta': 0.045},
        {'zeta': 4, 'lambda': -1, 'eta': 0.045},
        {'zeta': 4, 'lambda': 1, 'eta': 0.045},
        {'zeta': 16, 'lambda': -1, 'eta': 0.045},
        {'zeta': 16, 'lambda': 1, 'eta': 0.045},
        {'zeta': 1, 'lambda': -1, 'eta': 0.08},
        {'zeta': 1, 'lambda': 1, 'eta': 0.08},
        {'zeta': 2, 'lambda': -1, 'eta': 0.08},
        {'zeta': 2, 'lambda': 1, 'eta': 0.08},
        {'zeta': 4, 'lambda': -1, 'eta': 0.08},
        {'zeta': 4, 'lambda': 1, 'eta': 0.08},
        # {'zeta':16,  'lambda':-1,  'eta':0.08 },
        {'zeta': 16, 'lambda': 1, 'eta': 0.08},
    ]

    # tranfer units from bohr to angstrom
    bhor2ang = 0.529177
    for key, values in params.items():
        for val in values:
            if key == 'g2':
                val['eta'] /= bhor2ang ** 2
            elif key == 'g4':
                val['eta'] /= bhor2ang ** 2

    return params


def get_set31():
    """Hyperparameters for symmetry functions, as discussed in:
    Nongnuch Artrith and Jorg Behler. "High-dimensional neural network potentials
    for metal surfaces: A prototype study for copper." Physical Review B 85, no. 4
    (2012): 045439.
    """

    params = OrderedDict()

    params['g2'] = [
        {'eta': 0.0009, 'Rs': 0.0},
        {'eta': 0.01, 'Rs': 0.0},
        {'eta': 0.02, 'Rs': 0.0},
        {'eta': 0.035, 'Rs': 0.0},
        {'eta': 0.06, 'Rs': 0.0},
        {'eta': 0.1, 'Rs': 0.0},
        {'eta': 0.2, 'Rs': 0.0},
        {'eta': 0.4, 'Rs': 0.0},
    ]

    params['g4'] = [
        {'zeta': 1, 'lambda': -1, 'eta': 0.0001},
        {'zeta': 1, 'lambda': 1, 'eta': 0.0001},
        {'zeta': 2, 'lambda': -1, 'eta': 0.0001},
        {'zeta': 2, 'lambda': 1, 'eta': 0.0001},
        {'zeta': 1, 'lambda': -1, 'eta': 0.003},
        {'zeta': 1, 'lambda': 1, 'eta': 0.003},
        {'zeta': 2, 'lambda': -1, 'eta': 0.003},
        {'zeta': 2, 'lambda': 1, 'eta': 0.003},
        {'zeta': 1, 'lambda': 1, 'eta': 0.008},
        {'zeta': 2, 'lambda': 1, 'eta': 0.008},
        {'zeta': 1, 'lambda': 1, 'eta': 0.015},
        {'zeta': 2, 'lambda': 1, 'eta': 0.015},
        {'zeta': 4, 'lambda': 1, 'eta': 0.015},
        {'zeta': 16, 'lambda': 1, 'eta': 0.015},
        {'zeta': 1, 'lambda': 1, 'eta': 0.025},
        {'zeta': 2, 'lambda': 1, 'eta': 0.025},
        {'zeta': 4, 'lambda': 1, 'eta': 0.025},
        {'zeta': 16, 'lambda': 1, 'eta': 0.025},
        {'zeta': 1, 'lambda': 1, 'eta': 0.045},
        {'zeta': 2, 'lambda': 1, 'eta': 0.045},
        {'zeta': 4, 'lambda': 1, 'eta': 0.045},
        {'zeta': 16, 'lambda': 1, 'eta': 0.045},
    ]

    # tranfer units from bohr to angstrom
    bhor2ang = 0.529177
    for key, values in params.items():
        for val in values:
            if key == 'g2':
                val['eta'] /= bhor2ang ** 2
            elif key == 'g4':
                val['eta'] /= bhor2ang ** 2

    return params


class SymmetryFunctionError(Exception):
    def __init__(self, msg):
        super(SymmetryFunctionError, self).__init__(msg)
        self.msg = msg

    def __expr__(self):
        return self.msg



import os
import sys
import logging
import pickle
import numpy as np
import multiprocessing as mp
import kliff
from kliff import parallel
from kliff.atomic_data import atomic_number
from kliff.error import InputError

logger = kliff.logger.get_logger(__name__)


class Descriptor:
    """Base class of atomic enviroment descriptors.

    Preprocess dataset to generate fingerprints. This is the base class for all
    descriptors, so it should not be used directly. Instead, descriptors built on top
    of this such as :class:`~kliff.descriptors.SymmetryFunction` and
    :class:`~kliff.descriptors.Bispectrum` can be used to transform the atomic
    enviroment information into fingerprints.

    Parameters
    ----------
    cut_dists: dict
        Cutoff distances, with key of the form ``A-B`` where ``A`` and ``B`` are
        atomic species string, and value should be a float.

    cut_name: str
        Name of the cutoff function, such as ``cos``, ``P3``, ``P7``.

    hyperparams: dict
        A dictionary of the hyperparams of the descriptor.

    normalize: bool (optional)
        If ``True``, the fingerprints is centered and normalized according to:
        ``zeta = (zeta - mean(zeta)) / stdev(zeta)``

    dtype: np.dtype
        Data type for the generated fingerprints, such as ``np.float32`` and
        ``np.float64``.

    Attributes
    ----------
    size: int
        Lengh of the fingerprint vector.

    mean: list
        Mean of the fingerprints.

    stdev: list
        Standard deviation of the fingerprints.
    """

    def __init__(
        self, cut_dists, cut_name, hyperparams, normalize=True, dtype=np.float32
    ):

        self.cut_dists = cut_dists
        self.cut_name = cut_name
        self.hyperparams = hyperparams
        self.normalize = normalize
        self.dtype = dtype

        self.size = None
        self.mean = None
        self.stdev = None

    #    def set_cutoff(self, name, values):
    #        """Set the cutoff used in the descriptor.
    #
    #        Parameters
    #        ----------
    #        name: str
    #            Name of the cutoff, such as ``cos``, ``P3``, ``P7``.
    #
    #        values: dict
    #            Values for the cutoff, with key of the form ``A-B`` where ``A`` and ``B``
    #            are atomic species, and value should be a float.
    #
    #        Example
    #        -------
    #        >>> desc = Descriptor()
    #        >>> name = 'cos'
    #        >>> values = {'C-C':4.5,'H-H':3.0,'C-H':4.0}
    #        >>> desc.set_cutoff(name, values)
    #        """
    #
    #        self.cutname = name
    #        self.cutoff = generate_full_cutoff(values)
    #        self.species_code = dict()
    #
    #        species = get_species_from_cutoff(values)
    #        num_species = len(species)
    #
    #        rcutsym = np.zeros([num_species, num_species], dtype=np.double)
    #        try:
    #            for i, si in enumerate(species):
    #                self.species_code[si] = i
    #                for j, sj in enumerate(species):
    #                    rcutsym[i][j] = self.cutoff[si+'-'+sj]
    #        except KeyError as e:
    #            raise InputError('Cutoff for "{}" not provided.'.format(e))
    #
    #    def set_hyperparams(self, hyperparams):
    #        """Set the hyperparameters that the descriptor needs.
    #
    #        Parameters
    #        ----------
    #        hyperparams: dict
    #            Hyperparameters of the descriptor.
    #        """
    #        self.hyperparams = hyperparams

    def generate_train_fingerprints(
        self,
        configs,
        grad=False,
        reuse=False,
        prefix='fingerprints',
        nprocs=mp.cpu_count(),
    ):
        """Convert training set to fingerprints.

        Parameters
        ----------

        configs: list of Configuration objects

        grad: bool (optional)
            Whether to compute the gradient of fingerprints w.r.t. atomic coordinates.

        reuse: bool
            Whether to reuse the fingerprints if one is found at: {prefix}/train.pkl.

        prefix: str
            Directory name where the generated fingerprints are stored.
            Path to the generated fingerprints is: is: {prefix}/train.pkl

        nprocs: int
          Number of processors to be used.
        """

        fname = os.path.join(prefix, 'train.pkl')
        mean_stdev_name = os.path.join(prefix, 'mean_stdev.pkl')

        # TODO need to do additional check for the loaded data
        # restore data
        if os.path.exists(fname):
            logger.info('Found existing fingerprints: %s.', fname)
            if not reuse:
                os.remove(fname)
                logger.info(
                    'Delete existing fingerprints: %s, new ones is genereated.', fname
                )
            else:
                if self.normalize:
                    logger.info('Restore mean and stdev to: %s.', mean_stdev_name)
                    self.mean, self.stdev = load_mean_stdev(mean_stdev_name)
                return fname

        # generate data
        all_zeta, all_dzetadr = self.calc_zeta_dzetadr(configs, grad, nprocs)
        if self.normalize:
            if all_zeta is not None:
                stacked = np.concatenate(all_zeta)
                self.mean = np.mean(stacked, axis=0)
                self.stdev = np.std(stacked, axis=0)
            else:
                self.mean, self.stdev = self.welford_mean_and_stdev(configs, grad)
            dump_mean_stdev(self.mean, self.stdev, mean_stdev_name)
        self.dump_fingerprints(configs, fname, all_zeta, all_dzetadr, grad, nprocs)

        return fname

    def generate_test_fingerprints(
        self,
        configs,
        grad=False,
        reuse=False,
        prefix='fingerprints',
        train_prefix=None,
        nprocs=mp.cpu_count(),
    ):
        """Convert test set to fingerprints.

        Parameters
        ----------
        configs: list of Configuration objects

        grad: bool (optional)
            Whether to compute the gradient of fingerprints w.r.t. atomic coordinates.

        reuse: bool
            Whether to reuse the fingerprints if one is found at: {prefix}/test.pkl.

        prefix: str
            Directory name where the generated fingerprints are stored.
            Path to the generated fingerprints is: is: {prefix}/test.pkl

        train_prefix: str (optional)
            Directory name where the generated training fingerprints are stored.
            Defaults to `prefix`.
            This effects only when:
            1) `normalize` is `True` in the initializer; and
            2) self.genereate_train_fingerprints() has not been called,
            because in this case we need to find the mean and stdev of the fingerprints
            of the training set, which is saved as a file name `mean_stdev.pkl` in the
            derectory specified by `train_prefix`.

        nprocs: int
            Number of processors to be used.
        """

        if train_prefix is None:
            train_prefix = prefix

        fname = os.path.join(prefix, 'test.pkl')
        mean_stdev_name = os.path.join(train_prefix, 'mean_stdev.pkl')

        # TODO need to do additional check for the loaded data
        # restore data
        if os.path.exists(fname):
            logger.info('Found existing fingerprints: %s.', fname)
            if not reuse:
                os.remove(fname)
                logger.info(
                    'Delete existing fingerprints: %s, new ones is genereated.', fname
                )
            else:
                if self.normalize and (self.mean is None or self.stdev is None):
                    logger.info('Restore mean and stdev to: %s.', mean_stdev_name)
                    self.mean, self.stdev = load_mean_stdev(mean_stdev_name)
                return fname

        # generate data
        if self.normalize and (self.mean is None or self.stdev is None):
            raise DescriptorError(
                'Cannot proceed to genereate fingerprints for test set without '
                'that for training set generated when normalization is required. '
                'Try generating training set fingerprints first by calling: '
                'genereate_train_fingerprints().'
            )
        all_zeta, all_dzetadr = self.calc_zeta_dzetadr(configs, grad, nprocs)
        self.dump_fingerprints(configs, fname, all_zeta, all_dzetadr, grad, nprocs)

        return fname

    def dump_fingerprints(
        self, configs, fname, all_zeta, all_dzetadr, grad, nprocs=mp.cpu_count()
    ):

        logger.info('Pickle fingerprint images to: %s.', fname)

        dirname = os.path.dirname(fname)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        with open(fname, 'ab') as f:
            for i, conf in enumerate(configs):
                if i % 100 == 0:
                    logger.info('Processing configuration: %d.', i)
                zeta = all_zeta[i]
                dzetadr = all_dzetadr[i]
                if zeta is None:
                    zeta, dzetadr = self.transform(conf, grad=grad)

                # reshape dzeta from a 4D to a 3D array (combining the last two dims)
                if grad:
                    new_shape = (dzetadr.shape[0], dzetadr.shape[1], -1)
                    dzetadr = dzetadr.reshape(new_shape)

                # centering and normalization
                if self.normalize:
                    zeta = (zeta - self.mean) / self.stdev
                    if grad:
                        stdev_3d = np.atleast_3d(self.stdev)
                        dzetadr = dzetadr / stdev_3d

                # pickling data
                name = conf.get_identifier()
                species = conf.get_species()
                species = np.asarray([atomic_number[i] for i in species], np.intc)
                weight = np.asarray(conf.get_weight(), self.dtype)
                zeta = np.asarray(zeta, self.dtype)
                energy = np.asarray(conf.get_energy(), self.dtype)
                if grad:
                    dzetadr = np.asarray(dzetadr, self.dtype)
                    forces = np.asarray(conf.get_forces(), self.dtype)

                # TODO maybe change num atoms by species to species list or even conf
                example = {
                    'name': name,
                    'species': species,
                    'weight': weight,
                    'zeta': zeta,
                    'energy': energy,
                }
                if grad:
                    example['dzeta_dr'] = dzetadr
                    example['forces'] = forces
                pickle.dump(example, f)

        logger.info('Processing %d configurations finished.', len(configs))

    def calc_zeta_dzetadr(self, configs, grad, nprocs=mp.cpu_count()):
        try:
            rslt = parallel.parmap1(self.transform, configs, grad, nprocs=nprocs)
            all_zeta = [pair[0] for pair in rslt]
            all_dzetadr = [pair[1] for pair in rslt]
        except MemoryError as e:
            logger.info(
                '%s. Occurs in calculating fingerprints using parallel. '
                'Will fallback to use a serial version.',
                str(e),
            )
            all_zeta = [None for _ in len(configs)]
            all_dzetadr = [None for _ in len(configs)]
        return all_zeta, all_dzetadr

    def welford_mean_and_stdev(self, configs, grad):
        """Compute the mean and standard deviation of fingerprints.

        This running mean and standard method proposed by Welford is memory-efficient.
        Besides, it outperforms the naive method from suffering numerical instability
        for large dataset.

        see https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
        """

        # number of features
        conf = configs[0]
        zeta, _ = self.transform(conf, grad=grad)
        size = zeta.shape[1]

        # starting Welford's method
        n = 0
        mean = np.zeros(size)
        M2 = np.zeros(size)
        for i, conf in enumerate(configs):
            zeta, _ = transform(conf, grad=grad)
            for row in zeta:
                n += 1
                delta = row - mean
                mean += delta / n
                delta2 = row - mean
                M2 += delta * delta2
            if i % 100 == 0:
                sys.stdout.flush()
        stdev = np.sqrt(M2 / (n - 1))

        return mean, stdev

    def transform(self, conf, grad=False):
        """Transform atomic coords to atomic enviroment descriptor values.

        Parameters
        ----------
        conf: :class:`~kliff.dataset.Configuration` object
            A configuration of atoms.

        grad: bool (optional)
            Whether to compute the gradient of descriptor values w.r.t. atomic
            coordinates.

        Returns
        -------
        zeta: 2D array
            Descriptor values, each row for one atom.
            zeta has shape (num_atoms, num_descriptors), where num_atoms is the
            number of atoms in the configuration, and num_descriptors is the size
            of the descriptor vector (depending on the the choice of hyper-parameters).

        dzeta_dr: 4D array if grad is ``True``, otherwise ``None``
            Gradient of descriptor values w.r.t. atomic coordinates.
            dzeta_dr has shape (num_atoms, num_descriptors, num_atoms, DIM), where
            num_atoms and num_descriptors has the same meanings as described in zeta.
            DIM = 3 denotes three Cartesian coordinates.
        """

        raise NotImplementedError(
            'Method "transform" not implemented; it has to '
            'be needes to be added by any subclass.'
        )

    def get_size(self):
        """Retrun the size of the descritpor vector."""
        return self.size

    def get_mean(self):
        """Return a list of the mean of the fingerprints."""
        return self.mean.copy()

    def get_stdev(self):
        """Return a list of the standard deviation of the fingerprints."""
        return self.stdev.copy()

    def get_dtype(self):
        """Return the data type of the fingerprints."""
        return self.dtype

    def get_cutoff(self):
        """Return the name and values of cutoff. """
        return self.cut_name, self.cut_dists

    def get_hyperparams(self):
        """Return the hyperparameters of descriptors. """
        return self.hyperparams


def dump_mean_stdev(mean, stdev, fname):
    dirname = os.path.dirname(fname)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    data = {'mean': mean, 'stdev': stdev}
    with open(fname, 'wb') as f:
        pickle.dump(data, f)


def load_mean_stdev(fname):
    with open(fname, 'rb') as f:
        data = pickle.load(f)
        mean = data['mean']
        stdev = data['stdev']
    return mean, stdev


def load_fingerprints(fname):
    """Read preprocessed data.

    Parameters
    ----------
    fname: str
        Names of the pickled data file.

    fit_forces: bool
        Whether to fit to forces.

    Return
    ------
    Instance of tf.data.
    """
    data = []
    with open(fname, 'rb') as f:
        try:
            while True:
                x = pickle.load(f)
                data.append(x)
        except EOFError:
            pass
    return data


def generate_full_cutoff(cutoff):
    """Generate a full binary cutoff dictionary.

    For species pair `S1-S2` in the ``cutoff`` dictionary, add key `S2-S1` to it,
    whih the same value as `S1-S2`.

    Parameters
    ----------
    cutoff: dict
        Cutoff dictionary with key of the form ``A-B`` where ``A`` and ``B``
        are atomic species, and value should be a float.

    Return
    ------
    dict
        A dictionary with all combination of species as keys.

    Example
    -------
    >>> cutoff = {'C-C': 4.0, 'C-H':3.5}
    >>> generate_full_cutoff(cutoff)
        {'C-C': 4.0, 'C-H':3.5, 'H-C':3.5}
    """
    rcut2 = dict()
    for key, val in cutoff.items():
        s1, s2 = key.split('-')
        if s1 != s2:
            rcut2[s2 + '-' + s1] = val
    # merge
    rcut2.update(cutoff)

    return rcut2


def generate_species_code(cutoff):
    """Genereate species code info from cutoff dictionary.

    Parameters
    ----------
    cutoff: dict
        Cutoff dictionary with key of the form ``A-B`` where ``A`` and ``B``
        are atomic species, and value should be a float.

    Return
    ------
    species: dict
        A dictionary of species and the integer code, with keys the species in
        ``cutoff`` keys, and values integer code for species.

    Example
    -------
    >>> cutoff = {'C-C': 4.0, 'C-H':3.5}
    >>> generate_species_code(cutoff)
        {'C':0, 'H':1}
    """
    species = set()
    for key in cutoff:
        s1, s2 = key.split('-')
        species.update([s1, s2])
    species = list(species)
    species_code = {s: i for i, s in enumerate(species)}
    return species_code


class DescriptorError(Exception):
    def __init__(self, msg):
        super(DescriptorError, self).__init__(msg)
        self.msg = msg

    def __str__(self):
        return self.msg



from scipy.optimize import least_squares
from geodesiclm import geodesiclm
from cost import Cost
from kimcalculator import KIMcalculator
from modelparams import ModelParams
from keywords import InputKeywords
from dataset import DataSet
from lmplatconst import lmp_lat_const
from wrapcalculator import WrapCalculator
import sys
sys.path.append('../openkim_fit')
sys.path.append('../libs/geodesicLMv1.1/pythonInterface')


def test_endparse():

    # KIM model parameters
    modelname = 'Three_Body_Stillinger_Weber_MoS__MO_000000111111_001'
    params = ModelParams(modelname)
    fname = 'input/mos2_init_guess.txt'
    params.read(fname)
    params.echo_params()

    fname = 'input/data_4x4.edn'
    keys = ['unrelaxed-periodic-cell-vector-1',
            'unrelaxed-configuration-positions',
            'species']
    # key not in file
    #keys = [ 'haha']
    # list of list
    #keys = ['unrelaxed-configuration-positions']
    # list of one value
    #keys = ['unrelaxed-configuration-forces']
    # bare value
    #keys = ['unrelaxed-potential-energy']

    def func(arg):
        pass
    test_parse = WrapCalculator(params, fname, keys, func, modelname)
    rslt = test_parse.get_prediction()
    print rslt


# do a real fitting
def test_run_fitting():
    # KIM model parameters
    modelname = 'Three_Body_Stillinger_Weber_MoS__MO_000000111111_001'
    params = ModelParams(modelname)
    fname = 'input/mos2_init_guess.txt'
    params.read(fname)
    params.echo_params()

    # initial guess of params
    x0 = params.get_x0()
    print 'x0', x0

    # read config and reference data
    tset = DataSet()
    tset.read('training_set/training_set_MoS2.xyz')
    configs = tset.get_configs()

    # predictor
    func = lmp_lat_const
    fname = 'lattice_const.edn'
    keys = ['lattice-const']
    predictor = WrapCalculator(params, fname, keys, func, modelname)
    #rslt = predictor.get_prediction()
    # print 'init prediction', rslt

    # reference
    refs = [3.3]

    # cost function
    cst = Cost(params, nprocs=1)
    cst.add(predictor, refs, weight=1)

    cost = cst.get_cost(x0)
    print 'init cost', cost

    # optimize
    func = cst.get_residual
    #method = 'scipy-lm'
    method = 'geodesiclm'
    if method == 'geodesiclm':
        xf = geodesiclm(func, x0, h1=1e-5, h2=0.1, factoraccept=5, factorreject=2,
                        imethod=2, initialfactor=1, damp_model=0, print_level=2,
                        maxiters=10000, Cgoal=0.5e-7, artol=1.E-8, gtol=1.5e-10, xtol=1e-6,
                        xrtol=1.5e-6, ftol=-1, frtol=1.5e-6)
        print 'fitted params', xf

    elif method == 'scipy-lm':
        # test scipy.optimize minimization method
        res_1 = least_squares(func, x0, method='lm')
        print res_1

    params.echo_params()


if __name__ == '__main__':
    # test_endparse()
    test_run_fitting()
